/* aliases.h */
/* Last modified: April 1, 2006 */


#include <stdlib.h>

/* type definitions */
typedef unsigned long long BB;
typedef unsigned int ui;
typedef unsigned char uc;
typedef struct HashTable sHash;
typedef struct PV sPV;
typedef struct Others sOthers;
typedef struct Position sPos;
typedef struct PlyData sPlyData;
typedef struct TimeCheck sTC;


/* from fruit */
#if DEBUG
#  define ASSERT(a) { if (!(a)) { \
  printf("ERROR in %s line %d: \"" #a "\" failed\n",__FILE__,__LINE__); \
    exit(EXIT_FAILURE); }}
#else
#  define ASSERT(a)
#endif


#if !defined WINDOWS
/* common utility */
#define min(a,b)  (((a) > (b)) ? (b):(a))
#define max(a,b)  (((a) < (b)) ? (b):(a))
#endif

/* board utilities */
#define row(a) ((a) >> 3)
#define column(a) ((a) & 7)
#define rowDistance(f,t) (abs((int)(column(t)-column(f))))
#define columnDistance(f,t) (abs((int)(row(t)-row(f))))
#define distance(f,t) (max(rowDistance((f),(t)),columnDistance((f),(t))))

/* in move */
#define from(a) ((ui)((a) & 63))                                /*  6 bits */
#define to(a) ((ui)(((a) >> 6) & 63))                           /*  6 bits */
#define piece(a) ((ui)(((a) >> 12) & 63))                       /*  6 bits */
#define captured(a) ((ui)(((a) >> 18) & 63))                    /*  6 bits */
#define promotedTo(a) ((ui)(((a) >> 24) & 63))                  /*  6 bits */
#define score(a) ((int)((((a) >> 30) & 0xFFFF) - mateScore))    /* 16 bits */
#define draw(a) ((ui)(((a) >> 46) & 7))                         /*  3 bits */

/* for *Attacks() */
#define shiftToRow(B,i) ((B)>>(8*(i)))&255
/* Squares attacked by Rook at sq horizontally */
#define hRookAttacks(sq,pos) \
(RankAttack[(sq)][shiftToRow((pos)->WBoard | (pos)->BBoard, row((sq)))])
/* Squares attacked by Rook at sq vertically */
#define vRookAttacks(sq,pos) \
(RankAttack90[(sq)][shiftToRow((pos)->RotatedBoard90, column((sq)))])
/* Squares attacked by Rook at sq */
#define rookAttacks(sq,pos) \
((RankAttack[(sq)][shiftToRow((pos)->WBoard | \
    (pos)->BBoard, row((sq)))]) | \
 (RankAttack90[(sq)] \
    [shiftToRow((pos)->RotatedBoard90, column((sq)))]))
/* Squares attacked by Bishop at sq along left diagonal */
#define lBishopAttacks(sq,pos) \
(RankAttack45l[(sq)] \
   [shiftToRow((pos)->RotatedBoard45l, rank45l[(sq)])])
/* Squares attacked by Bishop at sq along right diagonal */
#define rBishopAttacks(sq,pos) \
(RankAttack45r[(sq)] \
   [shiftToRow((pos)->RotatedBoard45r, rank45r[(sq)])])
/* Squares attacked by Bishop at sq */
#define bishopAttacks(sq,pos) \
((RankAttack45l[(sq)] \
    [shiftToRow((pos)->RotatedBoard45l, rank45l[(sq)])]) | \
 (RankAttack45r[(sq)] \
    [shiftToRow((pos)->RotatedBoard45r, rank45r[(sq)])]))


/* for end game table base */
/*
  Service macro - initialize squares of the particular piece as well as
  counter for that piece. Note: dual initialization saves some time when
  TB is present, but waste for non-present TB. If we often will call
  probing function for an absent TB, maybe we shall split that code.
*/

#define  vInitSqCtr(rgCtr, rgSquares, i, bitboard) {             \
  int  cPieces=0;                                                \
  BB Temp=(bitboard);                                            \
  while (Temp) {                                                 \
    const squaret sq=firstOne (Temp);                            \
    (rgSquares)[(i)*C_PIECES+cPieces]=sq;                        \
    cPieces++;                                                   \
    Temp &= NilMask[sq];                                         \
  }                                                              \
  (rgCtr)[(i)]=cPieces;                                          \
}
