/* book.c */
/* Last modified: July 24, 2007 */


/* Read a PGN file, pick a game where our player wins, 
 * and put the first 20 moves into white or black opening book.
 * If the player is "allPlayers", then it takes all players' games. 
 */
void readPGN(char *file, char *player, int numberOfMoves)
{
  int i, j, iA = 0, iZ = 0;
  int inNumber = 1, inMove1 = 0, inMove2 = 0;
  int commentNo = 0, numberNo = 0, wOpeningNo = 0, bOpeningNo = 0;
  int maxBookMoves;
  int win = 13;                /* 1: 1-0, 0: 1/2-1/2, -1: 0-1 */
  int colorPlayed = 13;        /* white or black */
  int number[256];
  char c, comment[20][512], event[512];
  char line[512], move1[256][8], move2[256][8], numberStr[6];
  FILE *fdin, *fdout;

 /* How many openings are there in our opening books? */
  fdin = fopen("white", "r");
  while (fgets(line, 512, fdin))
    {
      if (strstr(line, "OpeningNo"))
	{
	  i = 0;  j = 12;
	  do 
	    {
	      numberStr[i++] = line[j++];
	    } while (line[j] != '"');

	  numberStr[i] = '\0';
	  wOpeningNo = atoi(numberStr);
	}
    }
  fclose(fdin);
  wOpeningNo++;

  fdin = fopen("black", "r");
  while (fgets(line, 512, fdin))
    {
      if (strstr(line, "OpeningNo"))
	{
	  i = 0;  j = 12;
	  do 
	    {
	      numberStr[i++] = line[j++];
	    } while (line[j] != '"');

	  numberStr[i] = '\0';
	  bOpeningNo = atoi(numberStr);
	}
    }
  fclose(fdin);
  bOpeningNo++;

  fdin = fopen(file, "r");
  while (fgets(line, 512, fdin))
    {
      if (line[0] == '[') 
	{
	  strcpy(comment[commentNo++], line);

	  if (strstr(line, "Event ") && commentNo == 1)
	    strcpy(event, line);
	  if (strstr(line, "Event ") && commentNo > 1)
	    {
	      /* write to opening book */
	      if (!strcmp(player, "allPlayers"))
		{
		  if (win == 1) colorPlayed = white;
		  else if (win == -1) colorPlayed = black;
		}
	      if ((win == 1 && colorPlayed == white) || 
		  (win == -1 && colorPlayed == black))
		{
		  if (win == 1) fdout = fopen("white", "a+");
		  else fdout = fopen("black", "a+");
		  
		  if (win == 1)
		    {
		      fprintf(fdout, "[OpeningNo \"%d\"]\n", wOpeningNo++);
		      fprintf(fdout, "[OpeningWeight 1]\n");
		    }
		  else
		    {
		      fprintf(fdout, "[OpeningNo \"%d\"]\n", bOpeningNo++);
		      fprintf(fdout, "[OpeningWeight 1]\n");
		    }
		  maxBookMoves = numberOfMoves;
		  if (numberNo < maxBookMoves) maxBookMoves = numberNo;
		  fprintf(fdout, "%s", event);
		  for (i = 0; i < commentNo-1; i++) 
		    fprintf(fdout, "%s", comment[i]);
		  for (i = 0; i < maxBookMoves; i++)
		    fprintf (fdout, "%d. %s %s\n", 
			     number[i], move1[i], move2[i]);
		  fprintf(fdout, "\n");
		  fclose(fdout);
		  strcpy(event, line);
		}
	      
	      /* initialize control variables for next opening */
	      inNumber = 1, inMove1 = inMove2 = 0;
	      commentNo = numberNo = 0;
	      win = colorPlayed = 13;
	    }
	  
	  /* who wins? */
	  if (strstr(line, "Result"))
	    {
	      if (strstr(line, "1-0")) win = 1;
	      else if (strstr(line, "1/2")) win = 0;
	      if (strstr(line, "0-1")) win = -1;
	    }

	  /* Does our player wins this game? */
	  if (strstr(line, player))
	    {
	      if (strstr(line, "White ")) colorPlayed = white;
	      else if (strstr(line, "Black ")) colorPlayed = black;
	    }
	}
      else 
	{
	  /* add a space to line so my algorithm below works as it is */
	  iZ = (int) strlen(line);
	  if (iZ > 1 && line[iZ-2] != ' ')/* the last character is a newline */
	    {
	      line[iZ-1] = ' ';
	      line[iZ] = '\n';
	      line[iZ+1] = '\0';
	    } 
	  
	  iA = iZ = 0;
	  while ((c = line[iZ++]) != '\0')
	    {
	      if (inNumber && (c == '.'))
		{
		  i = 0;
		  for (j = iA; j < iZ-1; j++) numberStr[i++] = line[j];
		  numberStr[i++] = '\0';
		  number[numberNo] = atoi(numberStr);
		  iA = iZ;
		  while ((c = line[iZ]) == ' ') iA = ++iZ;

		  inNumber = 0; inMove1 = 1;
		}
	      
	      if (inMove1 && (c == ' '))
		{
		  i = 0;
		  for (j = iA; j < iZ-1; j++) move1[numberNo][i++] = line[j];
		  move1[numberNo][i++] = '\0';
		  iA = iZ;
		  while ((c = line[iZ]) == ' ') iA = ++iZ;

		  inMove1 = 0; inMove2 = 1;
		}
	      
	      if (inMove2 && (c == ' '))
		{
		  i = 0;
		  for (j = iA; j < iZ-1; j++) move2[numberNo][i++] = line[j];
		  move2[numberNo][i++] = '\0';
		  iA = iZ;
		  while ((c = line[iZ]) == ' ') iA = ++iZ;

		  inMove2 = 0; inNumber = 1;
		  numberNo++;
		}
	    }
	}
    }     
  fclose(fdin);


  /* Take care of the last game because the file doesn't end with Event */
  /* write to opening book */
  if (!strcmp(player, "allPlayers"))
    {
      if (win == 1) colorPlayed = white;
      else if (win == -1) colorPlayed = black;
    }
  if ((win == 1 && colorPlayed == white) || 
      (win == -1 && colorPlayed == black))
    {
      if (win == 1) fdout = fopen("white", "a+");
      else fdout = fopen("black", "a+");
      
      if (win == 1)
	{
	  fprintf(fdout, "[OpeningNo \"%d\"]\n", wOpeningNo++);
	  fprintf(fdout, "[OpeningWeight 1]\n");
	}
      else
	{
	  fprintf(fdout, "[OpeningNo \"%d\"]\n", bOpeningNo++);
	  fprintf(fdout, "[OpeningWeight 1]\n");
	}
      maxBookMoves = numberOfMoves;
      if (numberNo < maxBookMoves) maxBookMoves = numberNo;
      fprintf(fdout, "%s", event);
      for (i = 0; i < commentNo-1; i++) 
	fprintf(fdout, "%s", comment[i]);
      for (i = 0; i < maxBookMoves; i++)
	fprintf (fdout, "%d. %s %s\n", 
		 number[i], move1[i], move2[i]);
      fprintf(fdout, "\n");
      fclose(fdout);
    }
}


/* Read a white file which contains games where white wins.  
 * Print out the FEN of a position 2 moves before white wins material.
 */
void whiteFEN(ui wToMove, sPos *pos, sPlyData *plyData, sHash **hash,
	      sOthers *others)
{
  int no = 0, number, printed = false;
  char line[512], numberStr[8], move1[8], move2[8], s[8], fen[2*maxMove][80];
  FILE *fdin;
  ui i, j, ply = 0, wCMoves, nCMoves, winMaterial = 0, wToMovePly;
  size_t k;
  BB winningCapture[32], nonCapture[128];
  BB theMove;
  int wPieceScore, bPieceScore, doubleBonus = 0;  /* not used */

  wToMovePly = wToMove;
  fdin = fopen("white", "r");
  while (fgets(line, 512, fdin) && ply < 2*maxMove)
    {
      if (line[0] == '[' || line[0] == '\n' || line[0] == '\r') 
	{
	  if (strstr(line, "OpeningNo"))
	    {
	      init(&wToMove, pos, plyData, hash, others);
	      ply = 0;
	      printed = false;

	      i = 0;  j = 12;
	      do 
		{
		  numberStr[i++] = line[j++];
		} while (line[j] != '"');

	      numberStr[i] = '\0';
	      no = atoi(numberStr);
	    }
	}
      else 
	{      
	  sscanf (line, "%d. %s %s", &number, move1, move2);

	  /* white's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData,
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move1); k > 1; k--)
		    if (move1[k] == '+' || move1[k] == '#') move1[k] = '\0';
		  if (!strcmp(move1, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData,
                                    others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move1); k > 1; k--)
			if (move1[k] == '+' || move1[k] == '#') 
			  move1[k] = '\0';
		      if (!strcmp(move1, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in white no %d: %d. %s\n",
		       no, number, move1); 
	      else
		printf("ERROR in black no %d: %d. %s\n",
		       no, number, move1); 
	      exit(EXIT_FAILURE);
	    }

	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */

	  /* black's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move2); k > 1; k--)
		    if (move2[k] == '+' || move2[k] == '#') move2[k] = '\0';
		  if (!strcmp(move2, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData,
				    others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move2); k > 1; k--)
			if (move2[k] == '+' || move2[k] == '#') 
			  move2[k] = '\0';
		      if (!strcmp(move2, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in white no %d: %d. ... %s\n",
		       no, number, move2); 
	      else
		printf("ERROR in black no %d: %d. ... %s\n",
		       no, number, move2); 
	      exit(EXIT_FAILURE);
	    }

	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */

	  /* save the FEN for this position */
	  getFEN(wToMove, pos, &plyData[ply], fen[ply]);
	  /* if white wins material for 3 moves, print FEN from 5 moves ago */
	  if (evalMaterial(pos, others, &wPieceScore, &bPieceScore,
			   &doubleBonus) 
	      >= 100) 
	    winMaterial++;
	  else winMaterial = 0;
	  /* 16 to avoid a gambit */ 
	  if (winMaterial == 3 && !printed && ply > 16+10)
	    {
	      printed = true;
	      printf("%s %d %d\n", fen[ply-10], no, number-5);
	    }
	}
    } 
  fclose(fdin);
}


/* Read a black file which contains games where black wins.  
 * Print out the FEN of a position 2 moves before black wins material.
 */
void blackFEN(ui wToMove, sPos *pos, sPlyData *plyData, sHash **hash, 
	      sOthers *others)
{
  int no = 0, number, printed = false;
  char line[512], numberStr[8], move1[8], move2[8], s[8], fen[2*maxMove][80];
  FILE *fdin;
  ui i, j, ply = 0, wCMoves, nCMoves, winMaterial = 0, wToMovePly;
  size_t k;
  BB winningCapture[32], nonCapture[128];
  BB theMove;
  int wPieceScore, bPieceScore, doubleBonus = 0;  /* not used */

  wToMovePly = wToMove;
  fdin = fopen("black", "r");
  while (fgets(line, 512, fdin) && ply < 2*maxMove)
    {
      if (line[0] == '[' || line[0] == '\n' || line[0] == '\r') 
	{
	  if (strstr(line, "OpeningNo"))
	    {
	      init(&wToMove, pos, plyData, hash, others);
	      ply = 0;
	      printed = false;

	      i = 0;  j = 12;
	      do 
		{
		  numberStr[i++] = line[j++];
		} while (line[j] != '"');

	      numberStr[i] = '\0';
	      no = atoi(numberStr);
	    }
	}
      else 
	{      
	  sscanf (line, "%d. %s %s", &number, move1, move2);

	  /* black's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move1); k > 1; k--)
		    if (move1[k] == '+' || move1[k] == '#') move1[k] = '\0';
		  if (!strcmp(move1, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData, others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move1); k > 1; k--)
			if (move1[k] == '+' || move1[k] == '#') 
			  move1[k] = '\0';
		      if (!strcmp(move1, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in black no %d: %d. %s\n",
		       no, number, move1); 
	      else
		printf("ERROR in black no %d: %d. %s\n",
		       no, number, move1); 
	      exit(EXIT_FAILURE);
	    }

	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */


	  /* save the FEN for this position */
	  getFEN(wToMove, pos, &plyData[ply], fen[ply]);
	  /* if black wins material for 3 moves, print FEN from 5 moves ago */
	  if (evalMaterial(pos, others, &wPieceScore, &bPieceScore,
			   &doubleBonus) 
	      <= -100) 
	    winMaterial++;
	  else winMaterial = 0;
	  /* 16 to avoid a gambit */ 
	  if (winMaterial == 3 && !printed && ply > 16+10)
	    {
	      printed = true;
	      printf("%s %d %d\n", fen[ply-10], no, number-5);
	    }

	  /* black's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move2); k > 1; k--)
		    if (move2[k] == '+' || move2[k] == '#') move2[k] = '\0';
		  if (!strcmp(move2, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData, others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move2); k > 1; k--)
			if (move2[k] == '+' || move2[k] == '#') 
			  move2[k] = '\0';
		      if (!strcmp(move2, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in black no %d: %d. ... %s\n",
		       no, number, move2); 
	      else
		printf("ERROR in black no %d: %d. ... %s\n",
		       no, number, move2); 
	      exit(EXIT_FAILURE);
	    }

	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */
	}
    } 
  fclose(fdin);
}


void readTextBooks(ui wToMove, sPos *pos, sPlyData *plyData, sHash **hash,
		   sOthers *others)
{
  int no = 0, count = 0;
  int number, done;
  char line[512], numberStr[8], move1[8], move2[8], s[8];
  FILE *fdin, *fout;
  ui i, j, ply = 0, wCMoves, nCMoves, posNo = 0, uniqueNo = 0, dot = 0;
  ui wToMovePly;
  size_t k;
  BB winningCapture[32], nonCapture[128];
  BB theMove;
  struct BookTable *bookPtr;
#ifdef WINDOWS
  double c0, c1;
#else
  struct timeval tv;
  struct timezone tz;
  long c0, c1;
#endif

#ifdef WINDOWS
  c0 = 1000000.0/CLOCKS_PER_SEC*clock();
#else
  gettimeofday(&tv, &tz);
  c0 = tv.tv_sec * 1000000;
#endif

  /* initialize book tables */
  memset((void *) wBookTable, 0, (bookSize+1)*sizeof(struct BookTable));
  memset((void *) bBookTable, 0, (bookSize+1)*sizeof(struct BookTable));

  printf("Reading opening books: ");

  /* white book */
  no = 0;
  wToMovePly = wToMove;
  fdin = fopen("white", "r");
  while (fgets(line, 512, fdin))
    {
      if (line[0] == '[' || line[0] == '\n' || line[0] == '\r') 
	{
	  if (strstr(line, "OpeningNo"))
	    {
	      init(&wToMove, pos, plyData, hash, others);
	      ply = 0;

	      i = 0;  j = 12;
	      do 
		{
		  numberStr[i++] = line[j++];
		} while (line[j] != '"');

	      numberStr[i] = '\0';
	      no = atoi(numberStr);
	    }
	  else if (strstr(line, "OpeningWeight"))
	    {
	      i = 0;  j = 15;
	      do 
		{
		  numberStr[i++] = line[j++];
		} while (line[j] != ']');

	      numberStr[i] = '\0';
	      count = atoi(numberStr);
	    }
	}
      else 
	{
	  sscanf (line, "%d. %s %s", &number, move1, move2);

	  /* white's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move1); k > 1; k--)
		    if (move1[k] == '+' || move1[k] == '#') move1[k] = '\0';
		  if (!strcmp(move1, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData, others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move1); k > 1; k--)
			if (move1[k] == '+' || move1[k] == '#') 
			  move1[k] = '\0';
		      if (!strcmp(move1, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in white no %d: %d. %s\n",
		       no, number, move1); 
	      else
		printf("ERROR in black no %d: %d. %s\n",
		       no, number, move1); 
	      exit(EXIT_FAILURE);
	    }

	  /* store in book hash table */
	  done = 0;
	  /* no rehashing, keep filling in the blank spaces */
	  for (bookPtr = &wBookTable[plyData[ply].key & (bookSize)];
	       bookPtr->count;
	       bookPtr++)
	    {
	      /* the move is already there */
	      if (bookPtr->key == (ui) plyData[ply].key &&
		  bookPtr->move == theMove)
		{
		  bookPtr->no = (unsigned short) no;
		  bookPtr->count += count;
		  done = 1;
		  break;
		}
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &wBookTable[bookSize])
		bookPtr = &wBookTable[0];
	    }
	  /* the move isn't there, but there's a blank space */
	  if (!done)
	    {
	      bookPtr->key = (ui) plyData[ply].key;
	      bookPtr->move = theMove;
	      bookPtr->no = (unsigned short) no;
	      bookPtr->count += count;
	      uniqueNo++;
	    }
	  
	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */

	  /* some statistics and nicities */
	  posNo++;  dot++;
	  if (dot > 65)
	    {
	      dot = 0;
	      printf("w");  fflush(stdout);
	    }

	  /* black's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move2); k > 1; k--)
		    if (move2[k] == '+' || move2[k] == '#') move2[k] = '\0';
		  if (!strcmp(move2, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData, others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move2); k > 1; k--)
			if (move2[k] == '+' || move2[k] == '#') 
			  move2[k] = '\0';
		      if (!strcmp(move2, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in white no %d: %d. ... %s\n",
		       no, number, move2); 
	      else
		printf("ERROR in black no %d: %d. ... %s\n",
		       no, number, move2); 
	      exit(EXIT_FAILURE);
	    }

	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */
	}
    }
  fclose(fdin);

  /* black book */
  no = 0;
  fdin = fopen("black", "r");
  while (fgets(line, 512, fdin))
    {
      if (line[0] == '[' || line[0] == '\n') 
	{
	  if (strstr(line, "OpeningNo"))
	    {
	      init(&wToMove, pos, plyData, hash, others);
	      ply = 0;

	      i = 0;  j = 12;
	      do 
		{
		  numberStr[i++] = line[j++];
		} while (line[j] != '"');

	      numberStr[i] = '\0';
	      no = atoi(numberStr);
	    }
	  else if (strstr(line, "OpeningWeight"))
	    {
	      i = 0;  j = 15;
	      do 
		{
		  numberStr[i++] = line[j++];
		} while (line[j] != ']');

	      numberStr[i] = '\0';
	      count = atoi(numberStr);
	    }
	}
      else 
	{
	  sscanf (line, "%d. %s %s", &number, move1, move2);

	  /* black's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move1); k > 1; k--)
		    if (move1[k] == '+' || move1[k] == '#') move1[k] = '\0';
		  if (!strcmp(move1, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos, plyData, others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move1); k > 1; k--)
			if (move1[k] == '+' || move1[k] == '#') 
			  move1[k] = '\0';
		      if (!strcmp(move1, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in black no %d: %d. %s\n",
		       no, number, move1); 
	      else
		printf("ERROR in black no %d: %d. %s\n",
		       no, number, move1); 
	      exit(EXIT_FAILURE);
	    }

	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */

	  /* black's move */
	  theMove = 0;
	  /* Try winning captures */
	  wCMoves = nCMoves = 0;
	  wCMoves = captures(wToMovePly, ply, winningCapture, pos,
			     plyData, others, 0);
	  if (wCMoves > 0) 
	    {
	      for (j = 0; j < wCMoves; j++)
		{
		  san(wToMovePly, ply, winningCapture[j], s, pos, plyData, 
		      others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(move2); k > 1; k--)
		    if (move2[k] == '+' || move2[k] == '#') move2[k] = '\0';
		  if (!strcmp(move2, s)) theMove = winningCapture[j];
		}
	    }
	  /* Try non captures */
	  if (theMove == 0)
	    {
	      nCMoves = nonCaptures(wToMovePly, ply, nonCapture, pos,
				    plyData, others, 0);
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(wToMovePly, ply, nonCapture[j], s, pos, plyData, 
			  others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(move2); k > 1; k--)
			if (move2[k] == '+' || move2[k] == '#') 
			  move2[k] = '\0';
		      if (!strcmp(move2, s)) theMove = nonCapture[j];
		    }
		}
	    }
	  if (theMove == 0)
	    {
	      if (wToMove)
		printf("ERROR in black no %d: %d. ... %s\n",
		       no, number, move2); 
	      else
		printf("ERROR in black no %d: %d. ... %s\n",
		       no, number, move2); 
	      exit(EXIT_FAILURE);
	    }

	  /* store in book hash table */
	  done = 0;
	  /* no rehashing, keep filling in the blank spaces */
	  for (bookPtr = &bBookTable[plyData[ply].key & (bookSize)];
	       bookPtr->count;
	       bookPtr++)
	    {
	      /* the move is already there */
	      if (bookPtr->key == (ui) plyData[ply].key &&
		  bookPtr->move == theMove)
		{
		  bookPtr->no = (unsigned short) no;
		  bookPtr->count += count;
		  done = 1;
		  break;
		}
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &bBookTable[bookSize])
		bookPtr = &bBookTable[0];
	    }
	  /* the move isn't there, but there's a blank space */
	  if (!done)
	    {
	      bookPtr->key = (ui) plyData[ply].key;
	      bookPtr->move = theMove;
	      bookPtr->no = (unsigned short) no;
	      bookPtr->count += count;
	      uniqueNo++;
	    }
	  
	  /* make the move, go to the next ply */
	  makeMove(wToMovePly, ++ply, theMove, pos, plyData, others);
	  wToMovePly ^= 1;
	  /* uncomment this if there's a problem with the book 
	     display(pos); */

	  /* some statistics and nicities */
	  posNo++;  dot++;
	  if (dot > 65)
	    {
	      dot = 0;
	      printf("b");  fflush(stdout);
	    }
	}
    }
  fclose(fdin);

#ifdef WINDOWS
  c1 = 1000000.0/CLOCKS_PER_SEC*clock();
#else
  gettimeofday(&tv, &tz);
  c1 = tv.tv_sec * 1000000;
#endif

  printf("\n%u positions in %g seconds\n", posNo, (float)(c1-c0)/1000000);
  printf("Unique positions: %2.2f percent of bookSize\n",
	 (posNo*100.0/(bookSize*2)));

  /* write the books to binary files */
  fout = fopen("white.bin", "w");
  fwrite(wBookTable, sizeof(struct BookTable), (bookSize+1), fout);
  fclose(fout);
  fout = fopen("black.bin", "w");
  fwrite(bBookTable, sizeof(struct BookTable), (bookSize+1), fout);
  fclose(fout);
}


void readBinaryBooks()
{
  FILE *fdin;

  /* read the books from binary files */
  if ((fdin = fopen("white.bin", "r")) != NULL)
    {
      fread(wBookTable, sizeof(struct BookTable), (bookSize+1), fdin);
      fclose(fdin);
    }
  else printf("Playing without white opening book.\n");
  if ((fdin = fopen("black.bin", "r")) != NULL)
    {
      fread(bBookTable, sizeof(struct BookTable), (bookSize+1), fdin);
      fclose(fdin);
    }
  else printf("Playing without black opening book.\n");
}


/* return a move from the opening book or zero */
BB openingBook(ui wToMove, char *usedWNo, char *usedBNo,
	       sPlyData *thisPlyData, sPos *pos)
{
  struct BookTable *bookPtr;
  int noOfMoves = 0, r;
  unsigned short count = 0;

  if (wToMove)
    {
      /* counting beans */
      for (bookPtr = &wBookTable[thisPlyData->key & (bookSize)];
	   bookPtr->count;
	   bookPtr++)
	{
	  if (bookPtr->key == (ui) thisPlyData->key)
	    {
	      if (usedWNo[bookPtr->no] == false) 
		{
		  noOfMoves++;
		  count += bookPtr->count;
		}
	      else printf("used opening: %d\n", bookPtr->no);
	    }
	  /* wrap around to the beginning of the table */
	  if (bookPtr == &wBookTable[bookSize])
	    bookPtr = &wBookTable[0];
	}

      /* picking a bean */
      if (noOfMoves == 0) return 0;
      else if (noOfMoves == 1) 
	{
	  for (bookPtr = &wBookTable[thisPlyData->key & (bookSize)];
	       bookPtr->count;
	       bookPtr++)
	    {
	      if (bookPtr->key == (ui) thisPlyData->key &&
		  usedWNo[bookPtr->no] == false)
		{
		  pos->openingNo = (ui) bookPtr->no;
		  return bookPtr->move;
		}
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &wBookTable[bookSize])
		bookPtr = &wBookTable[0];
	    }
	}
      else
	{
	  if (pos->Flags & OneMask[3]) r = (int) random32() % (count);
	  else r = 0;
	  for (bookPtr = &wBookTable[thisPlyData->key & (bookSize)];
	       bookPtr->count;
	       bookPtr++)
	    {
	      if (bookPtr->key == (ui) thisPlyData->key &&
		  usedWNo[bookPtr->no] == false)
		{
		  r -= bookPtr->count;
		  if (r < 0) 
		    {
		      pos->openingNo = (ui) bookPtr->no;
		      return bookPtr->move;
		    }
		}
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &wBookTable[bookSize])
		bookPtr = &wBookTable[0];
	    }
	}
    }
  else
    {
      /* counting beans */
      for (bookPtr = &bBookTable[thisPlyData->key & (bookSize)];
	   bookPtr->count;
	   bookPtr++)
	{
	  if (bookPtr->key == (ui) thisPlyData->key)
	    {
	      if (usedBNo[bookPtr->no] == false) 
		{
		  noOfMoves++;
		  count += bookPtr->count;
		}
	      else printf("used opening: %d\n", bookPtr->no);
	    }
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &bBookTable[bookSize])
		bookPtr = &bBookTable[0];
	}

      /* picking a bean */
      if (noOfMoves == 0) return 0;
      else if (noOfMoves == 1)
	{
	  for (bookPtr = &bBookTable[thisPlyData->key & (bookSize)];
	       bookPtr->count;
	       bookPtr++)
	    {
	      if (bookPtr->key == (ui) thisPlyData->key &&
		  usedBNo[bookPtr->no] == false)
		{
		  pos->openingNo = (ui) bookPtr->no;
		  return bookPtr->move;
		}
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &bBookTable[bookSize])
		bookPtr = &bBookTable[0];
	    }
	}
      else
	{
	  if (pos->Flags & OneMask[3]) r = (int) random32() % (count);
	  else r = 0;
	  for (bookPtr = &bBookTable[thisPlyData->key & (bookSize)];
	       bookPtr->count;
	       bookPtr++)
	    {
	      if (bookPtr->key == (ui) thisPlyData->key &&
		  usedBNo[bookPtr->no] == false)
		{
		  r -= bookPtr->count;
		  if (r < 0)
		    {
		      pos->openingNo = (ui) bookPtr->no;
		      return bookPtr->move;
		    }
		}
	      /* wrap around to the beginning of the table */
	      if (bookPtr == &bBookTable[bookSize])
		bookPtr = &bBookTable[0];
	    }
	}
    }
	  
  return 0;
}
