/* command.c */
/* Last modified: July 24, 2007 */


/* Get command, based on xboard's engine-intf.txt. */
#ifdef _REENTRANT
void *getCommand(void *arg)
#else
void getCommand()
#endif
{
  int i, commentLine;
  ui count;
  int wordNo;
  char lastC, c;

#ifdef _REENTRANT
  while (true)
    {
      usleep(10000);
#endif
      if ((go & goRest) || (go & goPonder))
	{
#ifdef NOT
	  if (!(pos->Flags & OneMask[0])) printf("What now? ");
#endif
	  
	  wordNo = -1;
	  count = 0;
	  lastC = ' ';
	  commentLine = false;
	  /* Better not assume that word[][] is always zero */
	  for (i = 0; i < WORDS; i++)
	    word[i][0] = '\0';
	  
	  c = getc(stdin);

	  do
	    {
	      /* The situation where getCommand() loops doing nothing */
	      if (wordNo == WORDS || count == LETTERS)
		{
		  wordNo = -1;       /* reset */
		  count = 0;
		  word[0][0] = '\0'; /* uncorrupt empty input from file */
#ifdef _REENTRANT
                  usleep(10000); /* don't loop millions of times a second */
#endif
		  continue;
		}

	      /* It's a comment, ignore the rest of the line. */
	      if ((wordNo == -1) && (count == 0) && (c == '#'))
		{
		  commentLine = true;
		}
	      
	      if (!commentLine)
		{
		  if (isspace((int) lastC) && isspace((int) c)) continue;
		  else if (isspace((int) lastC) && !(isspace((int) c)))
		    {
		      wordNo++;
		      count = 0;
		      word[wordNo][count++] = c;
		      lastC = c;
		    }
		  else if (!(isspace((int) lastC)) && (isspace((int) c)))
		    {
		      word[wordNo][count] = '\0';
		      /* Limit WORDS tokens per customer */
		      lastC = ' ';
		    }
		  else
		    {
		      word[wordNo][count++] = c;
		      lastC = c;
		    }
		}
	    }
	  while ((c=getc(stdin)) != newLine);
	  
	  /* Processing the command now */
	  if (wordNo >= 0)
	    {
	      word[wordNo][count] = '\0';

#ifdef MUTEX
	      pthread_mutex_lock(&goMutex);
#endif
	      go = goProcess;
#ifdef MUTEX
	      pthread_mutex_unlock(&goMutex);
#endif
#ifdef DEBUG
	      if (word[0][0] != '\0')
		for (i = 0; i <= wordNo; i++) 
		  printf("Word #%d is %s.\n", i, word[i]);
#endif
	    }
	}
#ifdef _REENTRANT
    }
#endif

#ifdef _REENTRANT
  /* This should not happen. */
  pthread_exit("Why do I exit out of a \"true loop\"?");
#endif
}


ui processCommand(char *word0, char *word1, char *word2, char *word3,
		  char *word4, char *word5, char *word6, char *word7,
		  char *word8, char *word9, char *word10, char *word11,
		  ui *wToMove, sPos *pos, sPlyData *plyData, sHash **hash, 
		  sOthers *others, ui acceptDraw, FILE *fd)
{
  ui i, j;
  ui ok = goProcess;
  char wordArray[WORDS][LETTERS], comment[256];
  int tempSize, reqSize;
  char fen[LETTERS];
  int score, wPieceScore, bPieceScore, doubleBonus = 0;
  ui wCMoves, nCMoves;
  BB winningCapture[32], nonCapture[128];
  BB theMove;
  char s[8];
  ui iAmSave;

  ASSERT(*wToMove == false || *wToMove == true);
  ASSERT(pos != NULL);
  ASSERT(plyData != NULL);
  ASSERT(others != NULL);
  ASSERT(acceptDraw == false || acceptDraw == true);
  ASSERT(fd != NULL);


  if (!strcasecmp(word0, "time"))
    {/*
       Set a clock that always belongs to the engine. N is a number in
       centiseconds (units of 1/100 second). Even if the engine changes
       to playing the opposite color, this clock remains with the 
       engine.
     */
      ok = goRest;
      pos->timeLeft = atof(word1);   /* in centi seconds */ 
    }
  else if (!strcasecmp(word0, "otim"))
    {/*
       Set a clock that always belongs to the opponent. N is a 
       number in centiseconds (units of 1/100 second). Even if the 
       opponent changes to playing the opposite color, this clock 
       remains with the opponent.
       
       If needed for purposes of board display in force mode (where
       the engine is not participating in the game) the time clock 
       should be associated with the last color that the engine was
       set to play, the otim clock with the opposite color.
       
       Beginning in protocol version 2, if you can't handle the 
       time and otim commands, you can use the "feature" command to
       disable them; see below.  The following techniques from 
       older protocol versions also work: You can ignore the time 
       and otim commands (that is, treat them as no-ops), or send 
       back "Error (unknown command): time" the first time you see 
       "time".
     */
      ok = goRest;
      pos->oTimeLeft = atof(word1);   /* in centi seconds */ 

      /* 
       * Correction for the delay in the way xboard is reporting otim
       * For example, in a 40 moves per 5 minutes game, when it is white's 
       * turn to move on Move 41, xboard reports otim without the additional
       * 5 minutes for the next 40 moves.  Hence, parrot has to add it.
       * WARNING: pos->moveNumber starts from 0, not 1 like the real move 
       * number.  It needs to start from 0 because it's used as an index of
       * gameListing[].
       */
      if (pos->movesPerTC > 0 &&
	  ((pos->moveNumber+1) % pos->movesPerTC == 0) &&
          pos->iAm == white && pos->moveNumber > 0)
        {
          pos->oTimeLeft += pos->timeLeft0;
        }
     }
  else if (!strcasecmp(word0, "white"))
    {/*
       (This command is obsolete as of protocol version 2, but is still
       sent in some situations to accommodate older engines unless you 
       disable it with the feature command.) Set White on move. Set the
       engine to play Black. Stop clocks.
     */
      ok = goRest;
      *wToMove = true;
      pos->iAm = black;
      /* in case I get "white go" */
      if (!strcasecmp(word1, "go"))
	{
	  pos->iAm ^= 1;
	  *wToMove = pos->iAm^1;
	  pos->Flags &= NilMask[1];   /* not forced */
	  ok = goSearch;
	}
    }
  else if (!strcasecmp(word0, "black"))
    {/*
       (This command is obsolete as of protocol version 2, but is still
       sent in some situations to accommodate older engines unless you 
       disable it with the feature command.) Set Black on move. Set the
       engine to play White. Stop clocks.
     */
      ok = goRest;
      *wToMove = false;
      pos->iAm = white;
      /* in case I get "black go" */
      if (!strcasecmp(word1, "go"))
	{
	  pos->iAm ^= 1;
	  *wToMove = pos->iAm^1;
	  pos->Flags &= NilMask[1];   /* not forced */
	  ok = goSearch  ;
	}
    }
  else if (!strcasecmp(word0, "result"))
    {/*
       After the end of each game, xboard will send you a result 
       command. You can use this command to trigger learning. RESULT is
       either 1-0, 0-1, 1/2-1/2, or *, indicating whether white won, 
       black won, the game was a draw, or the game was unfinished. The 
       COMMENT string is purely a human-readable comment; its content 
       is unspecified and subject to change. In ICS mode, it is passed 
       through from ICS uninterpreted.  Example:
       
       result 1-0 {White mates}
       
       Here are some notes on interpreting the "result" command. Some 
       apply only to playing on ICS ("Zippy" mode).
       
       If you won but did not just play a mate, your opponent must have
       resigned or forfeited. If you lost but were not just mated, you
       probably forfeited on time, or perhaps the operator resigned 
       manually.  If there was a draw for some nonobvious reason, 
       perhaps your opponent called your flag when he had insufficient 
       mating material (or vice versa), or perhaps the operator agreed 
       to a draw manually.
       
       You will get a result command even if you already know the game 
       ended -- for example, after you just checkmated your opponent. 
       In fact, if you send the "RESULT {COMMENT}" command (discussed 
       below), you will simply get the same thing fed back to you with 
       "result" tacked in front. You might not always get a "result *" 
       command, however. In particular, you won't get one in local 
       chess engine mode when the user stops playing by selecting 
       Reset, Edit Game, Exit or the like.
     */
      ok = goRest;
      j = 0;
      strcpy(wordArray[2], word2);
      strcpy(wordArray[3], word3);
      strcpy(wordArray[4], word4);
      strcpy(wordArray[5], word5);
      strcpy(wordArray[6], word6);
      strcpy(wordArray[7], word7);
      strcpy(wordArray[8], word8);
      strcpy(wordArray[9], word9);
      strcpy(wordArray[10], word10);
      strcpy(wordArray[11], word11);
      for (i = 2; i < WORDS; i++)
	{
	  if (strlen(wordArray[i]))
	    {
	      size_t k;
	      
	      for (k = 0; k < strlen(wordArray[i]); k++)
		comment[j++] = wordArray[i][k];
	      comment[j++] = ' ';
	    }
	}
      comment[j] = '\0';
      closing(fd, pos, word1, comment);
      init(wToMove, pos, plyData, hash, others);
    }
  else if (!strcasecmp(word0, "force"))
    {/*
       Set the engine to play neither color ("force mode").  Stop 
       clocks. The engine should check that moves received in force 
       mode are legal and made in the proper turn, but should not 
       think, ponder, or make move of its own.
     */
      ok = goRest;
      pos->Flags |= OneMask[1];   /* forced */
    }
  else if (!strcasecmp(word0, "go"))
    {/*
       Leave force mode and set the engine to play the color that is on
       move.  Associate the engine's clock with the color that is on 
       move, the opponent's clock with the color that is not on move. 
       Start the engine's clock. Start thinking and eventually make a 
       move.
     */
      ok = goRest;
      pos->iAm ^= 1;
      *wToMove = pos->iAm^1;
      pos->Flags &= NilMask[1];   /* not forced */
      ok = goSearch;
    }
  else if (!strcasecmp(word0, "level"))
    {/*
       level moversPerTC timeLeft0 timeInc
       timeLeft0 is in minutes
     */
      ok = goRest;
      pos->movesPerTC = (ui) atoi(word1);
      pos->timeLeft0 = atof(word2) * 6000; /* in centi seconds */ 
      pos->timeLeft = pos->timeLeft0;
      pos->timeInc = atof(word3) * 100;    /* in centi seconds */ 
    }
  else if (!strcasecmp(word0, "st"))
    {/*
       Set timeLeft in seconds.
       The commands "level" and "st" are not to be used together.
     */
      ok = goRest;
      pos->movesPerTC = 1;
      pos->timeLeft = pos->oTimeLeft = atof(word1) * 100;
      pos->timeInc = 0;                      /* in centi seconds */ 
    }
  else if (!strcasecmp(word0, "sd"))
    {/*
       The engine should limit its thinking to DEPTH ply.
     */
      ok = goRest;
      pos->searchDepth = (ui) atoi(word1);
    }
  else if (!strcasecmp(word0, "solution"))
    {/*
       Copy test position information to solution[] string.
     */
      ok = goRest;
      sprintf(pos->solution, "%s %s %s", word1, word2, word3);
    }
  else if (!strcasecmp(word0, "position"))
    {/*
       Read test position information.  Copy it to solution[] string.
       Do some initializations.
       WARNING: "position" has to be used before "setboard".
     */
      ok = goRest;
      init(wToMove, pos, plyData, hash, others);
      pos->Flags &= NilMask[3];    /* don't randomize */
      pos->Flags |= OneMask[6];    /* don't use opening book */
      pos->rOpening = 0;       /* don't evaluate base on a random opening */
#if !defined DEBUG
      pos->Flags &= NilMask[10];   /* not in timeReserved mode */
#endif
      sprintf(pos->solution, "%s %s %s", word1, word2, word3);
    }
  else if (!strcasecmp(word0, "printMove"))
    {/*
       The end of a position in testset
     */
      ok = goRest;
      printf("%s myMove %s\n", pos->solution, pos->myMove);
    }
  else if (!strcasecmp(word0, "setboard"))
    {/*
       The setboard command is the new way to set up positions, 
       beginning in protocol version 2. It is not used unless it has 
       been selected with the feature command. Here FEN is a position 
       in Forsythe-Edwards Notation, as defined in the PGN standard.
       
       Illegal positions: Note that either setboard or edit can be used
       to send an illegal position to the engine. The user can create 
       any position with xboard's Edit Position command (even, say, an 
       empty board, or a board with 64 white kings and no black ones). 
       If your engine receives a position that it considers illegal, I 
       suggest that you send the response "tellusererror Illegal 
       position", and then respond to any attempted move with "Illegal 
       move" until the next new, edit, or setboard command
     */
      ok = goRest;
      setBoard(word1, word2[0], word3, word4, word5, word6, wToMove, pos, 
	       &plyData[0], others);
      plyData[1] = plyData[0];
    }
  else if (!strcasecmp(word0, "getfen"))
    {
      ok = goRest;
      getFEN(*wToMove, pos, &plyData[0], fen);
      printf("%s\n", fen);
    }
  else if (!strcasecmp(word0, "reverse"))
    {
      ok = goRest;
      reverse(wToMove, pos, &plyData[0], others, fen);
      plyData[1] = plyData[0];
      printf("%s\n", fen);
    }
  else if (!strcasecmp(word0, "draw"))
    {/*
       The engine's opponent offers the engine a draw. To accept the 
       draw, send "offer draw". To decline, ignore the offer (that is, 
       send nothing". If you're playing on ICS, it's possible for the 
       draw offer to have been withdrawn by the time you accept it, so 
       don't assume the game is over because you accept a draw offer. 
       Continue playing until xboard tells you the game is over. See 
       also "offer draw" below.
     */

      /* acceptDraw is set in main() */
      if (acceptDraw == true)
	{
	  if (pos->Flags & OneMask[0]) printf("offer draw\n");
	}

      /* Is this position a draw by rule? */
      switch(isDrawByRule(0, pos, &plyData[0]))
	{
	case 0:
	  break;
	case 3:
	  if (pos->Flags & OneMask[0]) printf("offer draw\n");
	  break;
	case 13:
	  if (pos->Flags & OneMask[0]) printf("offer draw\n");
	  break;
	case 50:
	  if (pos->Flags & OneMask[0]) printf("offer draw\n");
	  break;
	default:
	  break;
	}
      ok = goRest;
    }
  else if (!strcasecmp(word0, "hard"))
    {/*
       Turn on pondering (thinking on the opponent's time, also known 
       as "permanent brain"). xboard will not make any assumption about
       what your default is for pondering or whether "new" affects this
       setting.
     */
      ok = goRest;
#if !defined HANDICAPPED_OPPONENT
      /* don't let xboard force you to turn on pondering in handicap mode */
      pos->Flags |= OneMask[11];     /* ponder on */
#endif
    }
  else if (!strcasecmp(word0, "easy"))
    {/*
       Turn off pondering.
     */
      ok = goRest;
#if !defined HANDICAPPED_OPPONENT
      /* don't let xboard force you to turn off pondering in handicap mode */
      pos->Flags &= NilMask[11];     /* ponder off */
#endif
    }
  else if (!strcasecmp(word0, "ponder"))
    {/* ponder on or off */
      ok = goRest;
      if (!strcasecmp(word1, "on")) pos->Flags |= OneMask[11];
      else if (!strcasecmp(word1, "off")) pos->Flags &= NilMask[11];
    }
  else if (!strcasecmp(word0, "new"))
    {/*
       Reset the board to the standard chess starting position. Set 
       White on move. Leave force mode and set the engine to play 
       Black. Associate the engine's clock with Black and the 
       opponent's clock with White. Reset clocks and time controls to 
       the start of a new game. Stop clocks. Do not ponder on this 
       move, even if pondering is on. Remove any search depth limit 
       previously set by the sd command.
     */
      ok = goRest;
      init(wToMove, pos, plyData, hash, others);
#if !defined TOURNAMENT
      printf("tellics seek 1 0 formula\n"); /* put out a challenge on FICS */  
      printf("tellics seek 2 12 formula\n");/* put out a challenge on FICS */  
      printf("tellics seek 28 1 formula\n");/* put out a challenge on FICS */  
      printf("tellics resume\n");  /* challenge opponents with stored games */
#endif
    }
  else if (!strcasecmp(word0, "rating"))
    {/*
       In ICS mode, xboard obtains the ICS opponent's rating from the
       "Creating:" message that appears before each game. (This message
       may not appear on servers using outdated versions of the FICS 
       code.) In Zippy mode, it sends these ratings on to the chess 
       engine using the "rating" command. The chess engine's own rating
       comes first, and if either opponent is not rated, his rating is 
       given as 0. In the future this command may also be used in other
       modes, if ratings are known.  Example:
       
       rating 2600 1500
     */
      ok = goRest;
      pos->deltaRating = atoi(word2) - atoi(word1);
    }
  else if (!strcasecmp(word0, "name"))
    {/*
       This command informs the engine of its opponent's name. When the
       engine is playing on a chess server, xboard obtains the 
       opponent's name from the server. When the engine is playing 
       locally against a human user, xboard obtains the user's login 
       name from the local operating system.  When the engine is 
       playing locally against another engine, xboard uses either the 
       other engine's filename or the name that the other engine 
       supplied in the myname option to the feature command. By 
       default, xboard uses the name command only when the engine is 
       playing on a chess server. Beginning in protocol version 2, you 
       can change this with the name option to the feature command; see
       below.
     */
      ok = goRest;
      strcpy(pos->opponentName, word1);
    }
  else if (!strcasecmp(word0, "computer"))
    {/*
       The opponent is also a computer chess engine. Some engines alter
       their playing style when they receive this command.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "post"))
    {/*
       Turn on thinking/pondering output. See Thinking Output section.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "nopost"))
    {/*
       Turn off thinking/pondering output.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "playother"))
    {/*
       (This command is new in protocol version 2. It is not sent 
       unless you enable it with the feature command.) Leave force mode
       and set the engine to play the color that is not on move. 
       Associate the opponent's clock with the color that is on move, 
       the engine's clock with the color that is not on move. Start the
       opponent's clock. If pondering is enabled, the engine should 
       begin pondering. If the engine later receives a move, it should 
       start thinking and eventually reply.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "?"))
    {/*
       Move now. If your engine is thinking, it should move 
       immediately; otherwise, the command should be ignored (treated 
       as a no-op"). It is permissible for your engine to always ignore
       the ? command. The only bad consequence is that xboard's Move 
       Now menu command will do nothing.
       
       It is also permissible for your engine to move immediately if it
       gets any command while thinking, as long as it processes the 
       command right after moving, but it's preferable if you don't do 
       this. For example, xboard may send post, nopost, easy, hard, 
       force, quit, or other commands while the engine is on move.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "exit"))
    {/*
       The chess engine should immediately exit. This command is used 
       when xboard is itself exiting, and also between games if the 
       -xreuse command line option is given (or -xreuse2 for the second
       engine"). See also Signals above.
     */
      exiting(fd, hash);
    }
  else if (!strcasecmp(word0, "quit"))
    {/*
       The chess engine should immediately exit. This command is used 
       when xboard is itself exiting, and also between games if the 
       -xreuse command line option is given (or -xreuse2 for the second
       engine"). See also Signals above.
     */
      exiting(fd, hash);
    }
  else if (!strcasecmp(word0, "protover"))
    {/*
       protover N
       Beginning in protocol version 2 (in which N=2"), this command 
       will be sent immediately after the "xboard" command. If you 
       receive some other command immediately after "xboard" (such as 
       "new""), you can assume that protocol version 1 is in use. The 
       "protover" command is the only new command that xboard always 
       sends in version 2. All other new commands to the engine are 
       sent only if the engine first enables them with the "feature" 
       command. Protocol versions will always be simple integers so 
       that they can easily be compared.
       
       Your engine should reply to the protover command by sending the
       "feature" command (see below) with the list of non-default 
       feature settings that you require, if any.
       
       Your engine should never refuse to run due to receiving a higher
       protocol version number than it is expecting! New protocol 
       versions will always be compatible with older ones by default; 
       the larger version number is simply a hui that additional 
       "feature" command options added in later protocol versions may 
       be accepted.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "accepted rejected"))
    {/*
       These commands may be sent to your engine in reply to the 
       "feature" command; see its documentation below.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "xboard"))
    {/*
       This command will be sent once immediately after your engine 
       process is started. You can use it to put your engine into 
       "xboard mode" if that is needed. If your engine prints a prompt 
       to ask for user input, you must turn off the prompt and output a
       newline when the "xboard" command comes in.
     */
      ok = goRest;
      setbuf(stdout, NULL);        /* Immediately write to stdout */
      /* Prevent xboard interrupt from killing parrotx */
      signal(SIGINT, SIG_IGN);
      signal (SIGTERM, SIG_IGN);
      pos->Flags |= OneMask[0];
    }
  else if (!strcasecmp(word0, "readPGN"))
    {/*
       readPGN PGNFileName playerName numberOfMoves
       Read a PGN file, pick a game where our player wins, and put the first 
       numberOfMoves moves into white or black opening book.  If the player
       is "allPlayers", then it takes all players' games. 
     */
      ok = goRest;
      readPGN(word1, word2, atoi(word3));
      init(wToMove, pos, plyData, hash, others);
    }
  else if (!strcasecmp(word0, "whiteFEN"))
    {/* 
       Read white file created by readPGN.  Print out the
       FEN of positions 2 plies before white wins material.
     */
      ok = goRest;
      whiteFEN(*wToMove, pos, plyData, hash, others);
    }
  else if (!strcasecmp(word0, "blackFEN"))
    {/* 
       Read black file created by readPGN.  Print out the
       FEN of positions 2 plies before black wins material.
     */
      ok = goRest;
      blackFEN(*wToMove, pos, plyData, hash, others);
    }
  else if (!strcasecmp(word0, "readTextBooks"))
    {/* 
       Read white and black files created by readPGN.  Process them 
       and save the openings in binary files white.bin and black.bin.
     */
      ok = goRest;
      readTextBooks(*wToMove, pos, plyData, hash, others);
      init(wToMove, pos, plyData, hash, others);
    }
  else if (!strcasecmp(word0, "bk"))
    {/*
       If the user selects "Book" from the xboard menu, xboard will 
       send your engine the command "bk". You can send any text you 
       like as the response, as long as each line begins with a blank 
       space or tab (\t character, and you send an empty line at the 
       end. The text pops up in a modal information dialog.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "pause resume"))
    {/*
       (These commands are new in protocol version 2 and will not be 
       sent unless feature pause=1 is set. At this writing, xboard 
       actually does not use the commands at all, but it or other 
       interfaces may use them in the future.) The "pause" command puts
       the engine into a special state where it does not think, ponder,
       or otherwise consume significant CPU time. The current thinking 
       or pondering (if any) is suspended and both player's clocks are 
       stopped. The only command that the interface may send to the 
       engine while it is in the paused state is "resume". The paused 
       thinking or pondering (if any) resumes from exactly where it 
       left off, and the clock of the player on move resumes running 
       from where it stopped.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "usermove MOVE"))
    {/*
       By default, moves are sent to the engine without a command name;
       the notation is just sent as a line by itself. Beginning in 
       protocol version 2, you can use the feature command to cause the
       command name "usermove" to be sent before the move. Example: 
       "usermove e2e4".
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "ping N"))
    {/*
       In this command, N is a decimal number. When you receive the 
       command, reply by sending the string pong N, where N is the same
       number you received. Important: You must not reply to a "ping" 
       command until you have finished executing all commands that you 
       received before it.  Pondering does not count; if you receive a 
       ping while pondering, you should reply immediately and continue 
       pondering. Because of the way xboard uses the ping command, if 
       you implement the other commands in this protocol, you should 
       never see a "ping" command when it is your move; however, if you
       do, you must not send the "pong" reply to xboard until after you
       send your move. For example, xboard may send "?" immediately 
       followed by "ping". If you implement the "?" command, you will 
       have moved by the time you see the subsequent ping command.  
       Similarly, xboard may send a sequence like "force", "new", 
       "ping". You must not send the pong response until after you have
       finished executing the "new" command and are ready for the new 
       game to start.
       
       The ping command is new in protocol version 2 and will not be 
       sent unless you enable it with the "feature" command. Its 
       purpose is to allow several race conditions that could occur in 
       previous versions of the protocol to be fixed, so it is highly 
       recommended that you implement it. It is especially important in
       simple engines that do not ponder and do not poll for input 
       while thinking, but it is needed in all engines.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "ics HOSTNAME"))
    {/*
       If HOSTNAME is "-", the engine is playing against a local 
       opponent; otherwise, the engine is playing on an Internet Chess 
       Server (ICS) with the given hostname. This command is new in 
       protocol version 2 and is not sent unless the engine has enabled
       it with the "feature" command.  Example: "ics freechess.org"
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "display"))
    {
      ok = goRest;
      display(pos);
    }
#ifdef UNDO_REMOVE
  else if (!strcasecmp(word0, "undo"))
    {/*
       If the user asks to back up one move, xboard will send you the 
       "undo" command. xboard will not send this command without 
       putting you in "force" mode first, so you don't have to worry 
       about what should happen if the user asks to undo a move your 
       engine made. (GNU Chess 4 actually switches to playing the 
       opposite color in this case.)
     */
      ok = goRest; 
      if (pos->repHead > 0) pos->repHead--;
      if (*wToMove == true) 
	{
	  pos->moveNumber--;
	  pos->failBook = gameListing[pos->moveNumber].wFailBook;
	  plyData[0] = gameListing[pos->moveNumber].bPlyData;
	  unmakeMove(*wToMove, 1, gameListing[pos->moveNumber].bMove,
		     pos, plyData);
	}
      else
	{
	  pos->failBook = gameListing[pos->moveNumber].bFailBook;
	  plyData[0] = gameListing[pos->moveNumber].wPlyData;
	  unmakeMove(*wToMove, 1, gameListing[pos->moveNumber].wMove,
		     pos, plyData);
	}
      pos->iAm ^= 1;
      display(pos);
    }
  else if (!strcasecmp(word0, "remove"))
    {/*
       If the user asks to retract a move, xboard will send you the 
       "remove" command. It sends this command only when the user is on
       move. Your engine should undo the last two moves (one for each 
       player) and continue playing the same color.
     */
      ok = goRest;

      /* undo the last move */
      if (pos->repHead > 0) pos->repHead--;
      if (*wToMove == true) 
	{
	  pos->moveNumber--;
	  pos->failBook = gameListing[pos->moveNumber].wFailBook;
	  plyData[0] = gameListing[pos->moveNumber].bPlyData;
	  unmakeMove(*wToMove, 1, gameListing[pos->moveNumber].bMove, pos, 
		     plyData);
	}
      else
	{
	  pos->failBook = gameListing[pos->moveNumber].bFailBook;
	  plyData[0] = gameListing[pos->moveNumber].wPlyData;
	  unmakeMove(*wToMove, 1, gameListing[pos->moveNumber].wMove, pos, 
		     plyData);
	}
      pos->iAm ^= 1;

      /* undo the move before the last move */
      if (pos->repHead > 0) pos->repHead--;
      if (*wToMove == true) 
	{
	  pos->moveNumber--;
	  plyData[0] = gameListing[pos->moveNumber].bPlyData;
	  unmakeMove(*wToMove, 1, gameListing[pos->moveNumber].bMove, pos, 
		     plyData);
	}
      else
	{
	  plyData[0] = gameListing[pos->moveNumber].wPlyData;
	  unmakeMove(*wToMove, 1, gameListing[pos->moveNumber].wMove, pos, 
		     plyData);
	}
      pos->iAm ^= 1;
      display(pos);
    }
#endif
  else if (!strcasecmp(word0, "hint"))
    {/*
       If the user asks for a hint, xboard sends your engine the 
       command "hint". Your engine should respond with "Hint: xxx", 
       where xxx is a suggested move. If there is no move to suggest, 
       you can ignore the hui command (that is, treat it as a no-op).
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "random"))
    {/*
       This command is specific to GNU Chess 4. You can either ignore 
       it completely (that is, treat it as a no-op) or implement it as 
       GNU Chess does. The command toggles "random" mode (that is, it 
       sets random = !random). In random mode, the engine adds a small 
       random value to its evaluation function to vary its play. The 
       "new" command sets random mode off.
     */
      /* "random on" */
      ok = goRest;
      if (!strcasecmp(word1, "on")) 
	{
	  pos->Flags |= OneMask[3];
	  /* choose an all purpose opening */
	  pos->rOpening = 1 + random32() % 3;
	}
      else if (!strcasecmp(word1, "off")) 
	{
	  pos->Flags &= NilMask[3];
	  /* don't evaluate base on a random opening */
	  pos->rOpening = 0;  
	}
    }
  else if (!strcasecmp(word0, "edit"))
    {/*
       The edit command is the old way to set up positions. For 
       compatibility  with old engines, it is still used by default, 
       but new engines may prefer to use the feature command (see 
       below) to cause xboard to use setboard instead. The edit command
       puts the chess engine into a special mode, where it accepts the 
       following subcommands:
       c                change current piece color, initially white
       Pa4 (for example place pawn of current color on a4
       xa4 (for example empty the square a4 (not used by xboard)
       #                clear board
       .                leave edit mode
       See the Idioms section below for additional subcommands used in
       ChessBase's implementation of the protocol.
       
       The edit command does not change the side to move. To set up a
       black-on-move position, xboard uses the following command 
       sequence:
       
       new
       force
       a2a3
       edit
       <edit commands>
       .
       
       This sequence is used to avoid the "black" command, which is now
       considered obsolete and which many engines never did implement 
       as specified in this document.
       
       After an edit command is complete, if a king and a rook are on 
       their home squares, castling is assumed to be available to them.
       En passant capture is assumed to be illegal on the current move 
       regardless of the positions of the pawns. The clock for the 50 
       move rule starts at zero, and for purposes of the draw by 
       repetition rule, no prior positions are deemed to have occurred.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "variant VARNAME"))
    {/*
       If the game is not standard chess, but a variant, this command 
       is sent after "new" and before the first move or "edit" command.
       Currently defined variant names are:
       
       wildcastle  Shuffle chess where king can castle from d file
       nocastle    Shuffle chess with no castling at all
       fischerandomFischer Random (not supported yet
       bughouse    Bughouse, ICC/FICS rules
       crazyhouse  Crazyhouse, ICC/FICS rules
       losers      Win by losing all pieces or getting mated (ICC)    
       suicide     Win by losing all pieces including king, or by 
       having fewer pieces when one player has no legal 
       moves (FICS)
       giveaway    Win by losing all pieces including king, or by 
       having no legal moves (ICC)
       twokings    Weird ICC wild 9
       kriegspiel  Kriegspiel (engines not supported
       atomic      Atomic
       3check      Win by giving check 3 times
       unknown     Unknown variant (not supported)
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "analyze"))
    {/*
       Enter analyze mode. See Analyze Mode section.
     */
      ok = goRest;
    }
  else if (!strcasecmp(word0, "score"))
    {
      /*
	Show the score of the side to move.
      */
      ok = goRest;
      iAmSave = pos->iAm;
      pos->iAm = *wToMove ^ 1;  /* wToMove = 1, iAm = white = 0 */
      score = evalMaterial(pos, others, &wPieceScore, &bPieceScore,
			   &doubleBonus);
      ASSERT(score >= -mateScore && score <= mateScore);
      if (score != drawScore)
	{
	  score = eval(*wToMove^1, 1, pos, plyData, others, score,
		       wPieceScore, bPieceScore, -mateScore, doubleBonus);
	  ASSERT(score >= -mateScore && score <= mateScore);
	}
      else score = 0;
      if (*wToMove == true) 
	printf("score = %d\n", score);
      else printf("score = %d\n", -score);
      pos->iAm = iAmSave;
    }
  else if (!strcasecmp(word0, "perft"))
    {
      /*
	Make moves to a certain depth to calculate how many legal positions 
	there are. 
      */
#ifdef WINDOWS
      double clock0, clock1;
#else
      struct timeval tv;
      struct timezone tz;
      BB clock0, clock1;
#endif
      ok = goRest;
      pos->sNodes = 0;

#ifdef WINDOWS      
      clock0 = 1000000.0/CLOCKS_PER_SEC*clock();
#else
      gettimeofday(&tv, &tz);
      clock0 = (BB) tv.tv_sec*1000000 + (BB) tv.tv_usec;
#endif
      perft(*wToMove, 0, (ui) atoi(word1), pos, plyData, others);
#ifdef WINDOWS      
      clock1 = 1000000.0/CLOCKS_PER_SEC*clock();
#else
      gettimeofday(&tv, &tz);
      clock1 = (BB) tv.tv_sec*1000000 + (BB) tv.tv_usec;
#endif
      printf("perft(%d) = %.0f (%0.2f knps)\n", atoi(word1), 
	     (double) pos->sNodes, 
	     1000.0*pos->sNodes/(clock1-clock0));
    }
  else if (!strcasecmp(word0, "hashsize"))
    {
      /*
	hashsize n where n, a power of 2, is hashsize in MBytes
      */
      ok = goRest;
      reqSize = atoi(word1);
      /* hashsize of 2^14-1 corresponds reqSize of 1 MBytes */
      tempSize = 1;
      for (i = 14; i < 32; i++)
	{
	  if (tempSize > reqSize) break;
	  else 
	    {
	      j = i;
	      tempSize *= 2;
	    }
	}
      pos->hashSize = 1;
      for (i = 1; i <= j; i++) pos->hashSize *= 2;
      pos->hashSize--;
    }
  else if (!strcasecmp(word0, "egtbcachesize"))
    {
      /* EGTB cache size, in MBytes */
      ok = goRest;
      pos->cacheSize = (ui)(atoi(word1)) * 1024 * 1024;
    }
  else if (!strcasecmp(word0, "egtbdir"))
    {/* EGTB directory */
      ok = goRest;
      sprintf(pos->egtbDir, "%s", word1);
    }
  else if (!strcasecmp(word0, "null"))
    {
      /*
	make a null move
      */
      ok = goRest;
      plyData[1] = plyData[0];
      *wToMove ^= 1;
      if (plyData[1].EP) 
	plyData[1].key ^= others->ePKey[firstOne(plyData[1].EP)];
      plyData[1].count50Moves = 0;
      plyData[1].EP = 0;

      /* Reset */
      plyData[0] = plyData[1];
      /* shift the killer moves */
      for (i = 0; i < maxPly-1; i++)
	{
	  pos->killerFrom[i] = pos->killerFrom[i+1];
	  pos->killerTo[i] = pos->killerTo[i+1];
	}
      display(pos);

      ok = goSearch;
    }
  else
    {/*
       MOVE
       See below for the syntax of moves. If the move is illegal, 
       print an error message; see the section "Commands from the 
       engine to xboard".  If the move is legal and in turn, make 
       it. If not in force mode, stop the opponent's clock, start 
       the engine's clock, start thinking, and eventually make a 
       move.
       
       When xboard sends your engine a move, it normally sends 
       coordinate algebraic notation. Examples:
       
       Normal moves:            e2e4
       Pawn promotion:          e7e8q
       Castling:                e1g1, e1c1, e8g8, e8c8
       Bughouse/crazyhouse drop:P@h3
       ICS Wild 0/1 castling:   d1f1, d1b1, d8f8, d8b8
       FischerRandom castling:  O-O, O-O-O (oh, not zero)
       
       Beginning in protocol version 2, you can use the feature 
       command to select SAN (standard algebraic notation) instead;
       for example, e4, Nf3, exd5, Bxf7+, Qxf7#, e8=Q, O-O, or 
       P@h3. Note that the last form, P@h3, is a extension to the 
       PGN standard's definition of SAN, which does not support 
       bughouse or crazyhouse.
       
       xboard doesn't reliably detect illegal moves, because it 
       does not keep track of castling unavailability due to king 
       or rook moves, or en passant availability. If xboard sends 
       an illegal move, send back an error message so that xboard 
       can retract it and inform the user; see the section 
       "Commands from the engine to xboard".
       
       e2e4, e7e8q, e1g1, e1c1, e8g8, e8c8, e4, exd5, or e8=Q
     */

      ok = goRest;
      theMove = 0;
      /* Try winning captures */
      wCMoves = nCMoves = 0;
      wCMoves = captures(*wToMove, 0, winningCapture, pos, 
			 plyData, others, 0);
      if (wCMoves > 0)
	{
	  for (j = 0; j < wCMoves; j++)
	    {
	      algebraic(winningCapture[j], s);
	      if (!strcmp(word0, s)) theMove = winningCapture[j];
	      else
		{
		  size_t k;
		  
		  san(*wToMove, 0, winningCapture[j], s, pos, plyData, others);
		  /* remove '+' or '#' */
		  for (k = strlen(s); k > 1; k--)
		    if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		  for (k = strlen(word0); k > 1; k--)
		    if (word0[k] == '+' || word0[k] == '#') 
		      word0[k] = '\0';
		  if (!strcmp(word0, s)) theMove = winningCapture[j];
		}
	    }
	}
      /* Try non captures */
      if (theMove == 0)
	{
	  nCMoves = nonCaptures(*wToMove, 0, nonCapture, pos, 
				plyData, others, 0);
	  if (nCMoves > 0) 
	    {
	      for (j = 0; j < nCMoves; j++)
		{
		  algebraic(nonCapture[j], s);
		  if (!strcmp(word0, s)) theMove = nonCapture[j];
		  else
		    {
		      size_t k;
		      
		      san(*wToMove, 0, nonCapture[j], s, pos, plyData, others);
		      /* remove '+' or '#' */
		      for (k = strlen(s); k > 1; k--)
			if (s[k] == '+' || s[k] == '#') s[k] = '\0';
		      for (k = strlen(word0); k > 1; k--)
			if (word0[k] == '+' || word0[k] == '#') 
			  word0[k] = '\0';
		      if (!strcmp(word0, s)) theMove = nonCapture[j];
		    }
		}
	    }
	}
      if (theMove == 0)
	{
	  if (word0[0] != '\0')
	    {
#ifdef DEBUG
	      display(pos);
	      printf("%d captures, %d non captures\n", wCMoves, nCMoves);
	      for (j = 0; j < wCMoves; j++)
		{
		  san(*wToMove, 0, winningCapture[j], s, pos, plyData, others);
		  printf("Winning capture #%d: %s\n", j, s);
		}
	      if (nCMoves > 0) 
		{
		  for (j = 0; j < nCMoves; j++)
		    {
		      san(*wToMove, 0, nonCapture[j], s, pos, plyData, others);
		      printf("Noncapture #%d: %s\n", j, s);
		    }
		}
#endif
	      printf("illegal move: %s\n", word0);
	    }
	}
      else
	{
	  /* write to the game listing */
	  san(*wToMove, 0, theMove, s, pos, plyData, others);
	  if (*wToMove == true)
	    {
	      if (pos->moveNumber == 0)
		{
		  /* listing headers */
		  fprintf(fd, "[Event \"%s\"]\n", pos->fullDate);
		  if (pos->iAm == white) fprintf(fd, "[White \"parrot\"] ");
		  else fprintf(fd, "[Black \"parrot\"] ");
		  fprintf(fd, "        score");
		  fprintf(fd, " opening");
		  fprintf(fd, " depth");
		  fprintf(fd, "\n");
		}
	      fprintf(fd, "white");
	    }
	  else fprintf(fd, "black");
	  fprintf(fd, "%4u.", (pos->moveNumber+1));
	  fprintf(fd, " %8s", s);
	  if (strlen(pos->solution) != 0) 
	    fprintf(fd, "\t                       %s", pos->solution); 
	  fprintf(fd, "\n");
	  fflush(fd);
	  
	  /* it's time to make the move now */
	  makeMove(*wToMove, 1, theMove, pos, plyData, others);
	  
	  /* Reset */
#ifdef UNDO_REMOVE
	  if (*wToMove == true)
	    {
	      gameListing[pos->moveNumber].wPlyData = plyData[0];
	      gameListing[pos->moveNumber].wFailBook = pos->failBook;
	    }
	  else
	    {
	      gameListing[pos->moveNumber].bPlyData = plyData[0];
	      gameListing[pos->moveNumber].bFailBook = pos->failBook;
	    }
#endif
	  plyData[0] = plyData[1];
	  /* shift the killer moves */
	  for (i = 0; i < maxPly-1; i++)
	    {
	      pos->killerFrom[i] = pos->killerFrom[i+1];
	      pos->killerTo[i] = pos->killerTo[i+1];
	    }
          *wToMove ^= 1;
	  if (*wToMove == true) pos->moveNumber++;
	  ASSERT(pos->moveNumber < maxMove);
	  if (plyData[0].count50Moves == 0) pos->repHead = 0;
	  else pos->repHead++;
	  display(pos);
	  
	  ok = goSearch;
	}
    }

  ASSERT(ok == goRest || ok == goSearch);
  return ok;
}
