/* eval.c */
/* Last modified: June 22, 2007 */


/********************** Leaf evaluation ***********************/
/* IMPORTANT: evalMaterial() returns white's score */
int evalMaterial(sPos *pos, sOthers *others, int *wPieceScore, 
		 int *bPieceScore, int *doubleBonus)
{
  int mScore, wPawnScore, bPawnScore;
  int pawnDiff, score0 = 0; 
  ui uselessMaterial, dontDouble;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);
  ASSERT(wPieceScore != NULL);
  ASSERT(bPieceScore != NULL);
  ASSERT(doubleBonus != NULL);

  /********************************************************************/
  /* material score                                                   */
  /********************************************************************/
  
  /* Don't forget to initialize them. */
  *wPieceScore = *bPieceScore = wPawnScore = bPawnScore = 0;

  /* non pawn white material */
  if (pos->WQBoard) *wPieceScore += bitCount(pos->WQBoard) * qValue;
  if (pos->WRBoard) *wPieceScore += bitCount(pos->WRBoard) * rValue;
  if (pos->WBBoard) *wPieceScore += bitCount(pos->WBBoard) * bValue;
  if (pos->WNBoard) *wPieceScore += bitCount(pos->WNBoard) * nValue;
  ASSERT(*wPieceScore <= mateScore);

  /* non pawn black material */
  if (pos->BQBoard) *bPieceScore -= bitCount(pos->BQBoard) * qValue;
  if (pos->BRBoard) *bPieceScore -= bitCount(pos->BRBoard) * rValue;
  if (pos->BBBoard) *bPieceScore -= bitCount(pos->BBBoard) * bValue;
  if (pos->BNBoard) *bPieceScore -= bitCount(pos->BNBoard) * nValue;
  ASSERT(*bPieceScore >= -mateScore);

  /* pawn material */
  if (pos->WPBoard) wPawnScore += bitCount(pos->WPBoard) * pValue;
  if (pos->BPBoard) bPawnScore -= bitCount(pos->BPBoard) * pValue;
  ASSERT(wPawnScore <= mateScore);
  ASSERT(bPawnScore >= -mateScore);

  /* Both sides do not have enough material to win */
  if (wPawnScore == 0 && bPawnScore == 0 && 
      *wPieceScore < rValue && *bPieceScore > -rValue)
    return drawScore;

  mScore = *wPieceScore + *bPieceScore + wPawnScore + bPawnScore;


  /* white bishop pair */
  if (bitCount(pos->WBBoard) == 2) score0 += 25;
  /* black bishop pair */
  if (bitCount(pos->BBBoard) == 2) score0 -= 25;

  pos->Flags &= NilMask[8];        /* clear drawish flag */
  /* rook endings are drawish */
  if (*wPieceScore == rValue && *bPieceScore == -rValue)
    {
      pos->Flags |= OneMask[8];
      pawnDiff = wPawnScore + bPawnScore;
      if (pawnDiff > 0) score0 -= 75;
      else if (pawnDiff < 0) score0 += 75;
    }
  /* Only bishops of opposite color left, drawish */
  else if (*wPieceScore == bValue && *bPieceScore == -bValue)
    {
      if ((pos->WBBoard & others->Dark) && 
	  (pos->BBBoard & others->Light))
	{
	  pos->Flags |= OneMask[8];
	  score0 -= mScore/2;
	}
      else if ((pos->WBBoard & others->Light) && 
	       (pos->BBBoard & others->Dark))
	{
	  pos->Flags |= OneMask[8];
	  score0 -= mScore/2;
	}
    }
  else
    {
      /* no pawn?  One minor piece advantage is pretty useless. */
      uselessMaterial = false;
      if (wPawnScore == 0)
	{
	  if (*wPieceScore < rValue && *wPieceScore > 0) 
	    {
	      score0 -= 225;
	      uselessMaterial = true;
	    }
	  else if (*bPieceScore == -rValue && 
		   (*wPieceScore + *bPieceScore) > 50 &&
		   (*wPieceScore + *bPieceScore) < rValue)
	    {
	      score0 -= (*wPieceScore + *bPieceScore) * 3/4;
	      uselessMaterial = true;
	    }
	}
      if (bPawnScore == 0)
	{
	  if (*bPieceScore > -rValue && *bPieceScore < 0) 
	    {
	      score0 += 225;
	      uselessMaterial = true;
	    }
	  else if (*wPieceScore == rValue && 
		   (*wPieceScore + *bPieceScore) < -50 &&
		   (*wPieceScore + *bPieceScore) > -rValue)
	    {
	      score0 -= (*bPieceScore + *wPieceScore) * 3/4;
	      uselessMaterial = true;
	    }
	}
      
      if (!uselessMaterial)
	{
	  dontDouble = false;
	  /* One bishop with one wrong rook pawn is pretty useless */
	  if ((*wPieceScore + wPawnScore == bValue + pValue) && mScore > 50)
	    {
	      if ((pos->WPBoard & others->Column[0]) && 
		  (pos->WBBoard & others->Dark))
		{
		  score0 = -*wPieceScore - wPawnScore + 50;
		  dontDouble = true;
		  pos->Flags |= OneMask[8];
		}
	      else if ((pos->WPBoard & others->Column[7]) && 
		       (pos->WBBoard & others->Light))
		{
		  score0 = -*wPieceScore - wPawnScore + 50;
		  dontDouble = true;
		  pos->Flags |= OneMask[8];
		}
	    }
	  if ((*bPieceScore + bPawnScore == -bValue - pValue) && mScore < -50)
	    {
	      if ((pos->BPBoard & others->Column[0]) && 
		  (pos->BBBoard & others->Light))
		{
		  score0 = -*wPieceScore - wPawnScore - 50;
		  dontDouble = true;
		  pos->Flags |= OneMask[8];
		}
	      else if ((pos->BPBoard & others->Column[7]) && 
		       (pos->BBBoard & others->Dark))
		{
		  score0 = -*wPieceScore - wPawnScore - 50;
		  dontDouble = true;
		  pos->Flags |= OneMask[8];
		}
	    }
	  
	  if (!dontDouble)
	    {
	      /* if non pawn material is ahead, 
		 trade queen to avoid complications */
	      if ((*wPieceScore + *bPieceScore > 0) && !pos->BQBoard)
		*doubleBonus = (*wPieceScore + *bPieceScore);
	      else if ((*wPieceScore + *bPieceScore < 0) && !pos->WQBoard)
		*doubleBonus = (*wPieceScore + *bPieceScore);
	    }
	}
    }

  return mScore + score0 + *doubleBonus;
}


#ifdef NOT
/* no potential passed pawn */
ui noPotentialPPawn(sPos *pos)
{
  ui i;

  ASSERT(pos != NULL);

  for (i = 0; i < 8; i++)
    {
      if (pos->wPCount[i] == 0)
	{
	  if (pos->bPCount[i] > 0) return true;
	}
      else if (pos->bPCount[i] == 0) return true;
      ASSERT((pos->wPCount[i] > 0 && pos->bPCount[i] > 0) ||
	     (pos->wPCount[i] == 0 && pos->bPCount[i] == 0));
    }
  return false;
}
#endif


int bonusFor7thRank(ui side, ui rank, ui fyle, sPos *pos)
{
  ASSERT(side == white || side == black);
  ASSERT(rank >= 0 && rank <= 7);
  ASSERT(fyle >= 0 && fyle <= 7);
  ASSERT(pos != NULL);

  if (side == white)
    {
      if ((rank == 6 && (pos->bKSq > H7)) ||
	  (rank == 1 && (pos->bKSq < A2)) ||
	  (fyle == 1 && (column(pos->bKSq) == 0)) ||
	  (fyle == 6 && (column(pos->bKSq) == 7)))
	return true;
    }
  else
    {
      if ((rank == 1 && (pos->wKSq < A2)) ||
	  (rank == 6 && (pos->wKSq > H7)) ||
	  (fyle == 1 && (column(pos->wKSq) == 0)) ||
	  (fyle == 6 && (column(pos->wKSq) == 7)))
	return true;
    }

  return false;
} 


static int wPBonus(ui wSleepy, sPos *pos, sOthers *others) 
{
  int score = 0;
  ui sq, rank, fyle, wKFyle, bKFyle, ourPawns, theirPawns;
  BB Temp;
  ui i, n;

  ASSERT(wSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White pawn score                                            */
  /********************************************************************/

  wKFyle = column(pos->wKSq);
  bKFyle = column(pos->bKSq);

  for (i = 0; i < 8; i++)
    {
      n = pos->wPCount[i];
      if (n)
	{
	  /* doubled or tripled or .... pawns */
	  if (n > 1) score -= (n-1) * PENALTY00;

	  if (wSleepy == 0)
	    {
	      sq = lastOne(pos->WPBoard & others->Column[i]);
	      ASSERT(sq >= A1 && sq <= H8);

	      /* pawn storm */
	      if (i < 4 && wKFyle > 3)         /* not in front of own king */
		{
		  rank = row(sq);
		  score += rank * BONUS17;
		  if (bKFyle < 4 && pos->WQBoard) score += rank * BONUS18;
		}
	      else if (i > 3 && wKFyle < 4)    /* not in front of own king */
		{
		  rank = row(sq);
		  score += rank * BONUS17;
		  if (bKFyle > 3 && pos->WQBoard) score += rank * BONUS18;
		}

	      /* passed pawns */
	      if (!(pos->BPBoard & others->Up8Of[sq]))
		{
		  theirPawns = 
		    bitCount(pos->BPBoard & others->Up8SideOf[sq]);
		  ASSERT(theirPawns <= 8);
		  if (!theirPawns) 
		    score += others->wPassedPSqvalue[sq];
		  else
		    {
		      ourPawns = bitCount(pos->WPBoard & 
					  (others->SideOf[sq] | 
					   others->Down8SideOf[sq]));
		      ASSERT(ourPawns <= 8);
		      if (ourPawns >= theirPawns)
			score += others->wPassedPSqvalue[sq];
		    }
		}
	    }
	}
    }
    
  Temp = pos->WPBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      score += others->wPSqvalue[sq];
      /* isolani */
      fyle = column(sq);
      if ((fyle == 0 || pos->wPCount[fyle-1] == 0) &&
	  (fyle == 7 || pos->wPCount[fyle+1] == 0))
	score -= PENALTY01;

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score > PBonus) PBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bPBonus(ui bSleepy, sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq, rank, fyle, wKFyle, bKFyle, ourPawns, theirPawns;
  BB Temp;
  ui i, n;

  ASSERT(bSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black pawn score                                            */
  /********************************************************************/

  wKFyle = column(pos->wKSq);
  bKFyle = column(pos->bKSq);

  for (i = 0; i < 8; i++)
    {
      n = pos->bPCount[i];
      if (n)
	{
	  /* doubled or tripled or .... pawns */
	  if (n > 1) score += (n-1) * PENALTY00;

	  if (bSleepy == 0)
	    {
	      sq = firstOne(pos->BPBoard & others->Column[i]);
	      ASSERT(sq >= A1 && sq <= H8);
	      
	      /* pawn storm */
	      if (i < 4 && bKFyle > 3)         /* not in front of own king */
		{
		  rank = 7-row(sq);
		  score -= rank * BONUS17;
		  if (wKFyle < 4 && pos->BQBoard) score -= rank * BONUS18;
		}
	      else if (i > 3 && bKFyle < 4)    /* not in front of own king */
		{
		  rank = 7-row(sq);
		  score -= rank * BONUS17;
		  if (wKFyle > 3 && pos->BQBoard) score -= rank * BONUS18;
		}

	      /* passed pawns */
	      if (!(pos->WPBoard & others->Down8Of[sq]))
		{
		  theirPawns = 
		    bitCount(pos->WPBoard & others->Down8SideOf[sq]);
		  ASSERT(theirPawns <= 8);
		  if (!theirPawns) 
		    score -= others->bPassedPSqvalue[sq];
		  else
		    {
		      ourPawns = bitCount(pos->BPBoard & 
					  (others->SideOf[sq] | 
					   others->Up8SideOf[sq]));
		      ASSERT(ourPawns <= 8);
		      if (ourPawns >= theirPawns) 
			score -= others->bPassedPSqvalue[sq];
		    }
		}
	    }
	}
    }
  
  Temp = pos->BPBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      score -= others->bPSqvalue[sq];
      /* isolani */
      fyle = column(sq);
      if ((fyle == 0 || pos->bPCount[fyle-1] == 0) &&
	  (fyle == 7 || pos->bPCount[fyle+1] == 0))
	  score += PENALTY01;

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wNBonus(ui wSleepy, sPos *pos, sOthers *others) 
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(wSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White knight score                                               */
  /********************************************************************/
  Temp = pos->WNBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      score += others->wNSqvalue[sq];

      AttackBoard = others->NMoves[sq] & ~pos->WBoard;
      score += bitCount(AttackBoard) * BONUS00;
      ASSERT(score <= mateScore);

      if (wSleepy == 0)
	{
	  /* Knight outpost */
	  if (others->BPCaptures[sq] & pos->WPBoard)
	    {
	      if (sq == D5)
		{
		  if (bitCount(pos->BNBoard) == 0 && 
		      !(pos->BBBoard & others->Light) &&
		      !(pos->BPBoard & (OneMask[C7] | OneMask[E7]))) 
		    score += BONUS01;
		  else score += BONUS02;
		}
	      else if (sq == E5)
		{
		  if (bitCount(pos->BNBoard) == 0 && 
		      !(pos->BBBoard & others->Dark) &&
		      !(pos->BPBoard & (OneMask[D7] | OneMask[F7]))) 
		    score += BONUS01;
		  else score += BONUS02;
		}
	      else if (sq == D6)
		{
		  if (bitCount(pos->BNBoard) == 0 && 
		      !(pos->BBBoard & others->Dark)) 
		    score += BONUS03;
		  else score += BONUS04;
		}
	      else if (sq == E6)
		{
		  if (bitCount(pos->BNBoard) == 0 && 
		      !(pos->BBBoard & others->Light)) 
		    score += BONUS03;
		  else score += BONUS04;
		}
	    }
	}

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score > NBonus) NBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bNBonus(ui bSleepy, sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(bSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black knight score                                               */
  /********************************************************************/
  Temp = pos->BNBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      score -= others->bNSqvalue[sq];

      AttackBoard = others->NMoves[sq] & ~pos->BBoard;
      score -= bitCount(AttackBoard) * BONUS00;
      ASSERT(score >= -mateScore);

      if (bSleepy == 0)
	{
	  /* Knight outpost */
	  if (others->WPCaptures[sq] & pos->BPBoard)
	    {
	      if (sq == D4)
		{
		  if (bitCount(pos->WNBoard) == 0 && 
		      !(pos->WBBoard & others->Dark) &&
		      !(pos->WPBoard & (OneMask[C2] | OneMask[E2]))) 
		    score -= BONUS01;
		  else score -= BONUS02;
		}
	      else if (sq == E4)
		{
		  if (bitCount(pos->WNBoard) == 0 && 
		      !(pos->WBBoard & others->Light) &&
		      !(pos->WPBoard & (OneMask[D2] | OneMask[F2]))) 
		    score -= BONUS01;
		  else score -= BONUS02;
		}
	      else if (sq == D3)
		{
		  if (bitCount(pos->WNBoard) == 0 && 
		      !(pos->WBBoard & others->Light)) 
		    score -= BONUS03;
		  else score -= BONUS04;
		}
	      else if (sq == E3)
		{
		  if (bitCount(pos->WNBoard) == 0 && 
		      !(pos->WBBoard & others->Dark)) 
		    score -= BONUS03;
		  else score -= BONUS04;
		}
	    }
	}

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wBBonus(sPos *pos, sOthers *others) 
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White bishop score                                               */
  /********************************************************************/
  Temp = pos->WBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      score += others->wBSqvalue[sq];

      AttackBoard = bishopAttacks(sq, pos) & ~pos->WBoard;
      score += bitCount(AttackBoard) * BONUS05;
      ASSERT(score <= mateScore);

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score > BBonus) BBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wBPenalty(sPos *pos) 
{
  int score = 0;
  ui sq;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* White bishop score                                               */
  /********************************************************************/
  Temp = pos->WBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);

      if (sq == D3 && (pos->WPBoard & OneMask[D2])) score -= PENALTY02;
      else if (sq == E3 && (pos->WPBoard & OneMask[E2])) score -= PENALTY02; 

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score < BPenalty) BPenalty = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bBBonus(sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black bishop score                                               */
  /********************************************************************/
  Temp = pos->BBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      score -= others->bBSqvalue[sq];

      AttackBoard = bishopAttacks(sq, pos) & ~pos->BBoard;
      score -= bitCount(AttackBoard) * BONUS05;
      ASSERT(score >= -mateScore);

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bBPenalty(sPos *pos)
{
  int score = 0;
  ui sq;
  BB Temp;


  /********************************************************************/
  /* Black bishop score                                               */
  /********************************************************************/
  Temp = pos->BBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);

      if (sq == D6 && (pos->BPBoard & OneMask[D7])) score += PENALTY02; 
      else if (sq == E6 && (pos->BPBoard & OneMask[E7])) score += PENALTY02; 

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wRBonus(ui wSleepy, int minPieceScore, sPos *pos, sOthers *others) 
{
  int score = 0;
  ui sq, fyle, rank;
  BB Temp;

  ASSERT(wSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White rook score                                                 */
  /********************************************************************/
  Temp = pos->WRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      fyle = column(sq);
      rank = row(sq);
      ASSERT(fyle < 8 && rank < 8);
      score += others->wRSqvalue[sq];

      /* semi open or open file */
      if (!pos->wPCount[fyle])
	{
	  score += BONUS06;
	  if (!pos->bPCount[fyle]) score += BONUS15;
	}

      if (wSleepy == 0 || minPieceScore <= middleGameValue)
	{
	  /* Rook in 7th rank */
	  if ((bitCount(pos->WRBoard | pos->WQBoard) > 1) &&
	      bonusFor7thRank(white, rank, fyle, pos))
	    score += BONUS07;
	}

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score > RBonus) RBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wRPenalty(sPos *pos)
{
  int score = 0;
  ui sq, fyle, rank;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* White rook score                                                 */
  /********************************************************************/
  Temp = pos->WRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      fyle = column(sq);
      rank = row(sq);
      ASSERT(fyle < 8 && rank < 8);

      /* Rook obstructed by king */
      if (row(pos->wKSq) == rank && rank < 2)
	{
	  if (fyle < 2 && column(pos->wKSq) < 3 &&
	      column(pos->wKSq) > fyle)
	    score -= PENALTY03;
	  else if (fyle > 5 && column(pos->wKSq) < fyle &&
		   column(pos->wKSq) > 4)
	    score -= PENALTY03;
	}

      /* Rook obstructed by pawn */
      if (fyle == 0 && 
	  ((pos->WPBoard & OneMask[sq+1]) ||
	   (pos->WPBoard & OneMask[sq+2])))
	score -= PENALTY04;
      else if (fyle == 7 && 
	       ((pos->WPBoard & OneMask[sq-1]) ||
		(pos->WPBoard & OneMask[sq-2])))
	score -= PENALTY04;

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score < RPenalty) RPenalty = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bRBonus(ui bSleepy, int minPieceScore, sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq, fyle, rank;
  BB Temp;

  ASSERT(bSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black rook score                                                 */
  /********************************************************************/
  Temp = pos->BRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      fyle = column(sq);
      rank = row(sq);
      ASSERT(fyle < 8 && rank < 8);
      score -= others->bRSqvalue[sq];

      /* semi open or open file */
      if (!pos->bPCount[fyle])
	{
	  score -= BONUS06;
	  if (!pos->wPCount[fyle]) score -= BONUS15;
	}

      if (bSleepy == 0 || minPieceScore <= middleGameValue)
	{
	  /* Rook in 7th rank */
	  if ((bitCount(pos->BRBoard | pos->BQBoard) > 1) &&
	      bonusFor7thRank(black, rank, fyle, pos))
	    score -= BONUS07;
	}

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bRPenalty(sPos *pos)
{
  int score = 0;
  ui sq, fyle, rank;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* Black rook score                                                 */
  /********************************************************************/
  Temp = pos->BRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      fyle = column(sq);
      rank = row(sq);
      ASSERT(fyle < 8 && rank < 8);

      /* Rook obstructed by king */
      if (row(pos->bKSq) == rank && rank > 5)
	{
	  if (fyle < 2 && column(pos->bKSq) < 3 &&
	      column(pos->bKSq) > fyle)
	    score += PENALTY03;
	  else if (fyle > 5 && column(pos->bKSq) < fyle &&
		   column(pos->bKSq) > 4)
	    score += PENALTY03;
	}

      /* Rook obstructed by pawn */
      if (fyle == 0 && 
	  ((pos->BPBoard & OneMask[sq+1]) ||
	   (pos->BPBoard & OneMask[sq+2])))
	score += PENALTY04;
      else if (fyle == 7 && 
	       ((pos->BPBoard & OneMask[sq-1]) ||
		(pos->BPBoard & OneMask[sq-2])))
	score += PENALTY04;

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wQBonus(ui wSleepy, int minPieceScore, sPos *pos, sOthers *others) 
{
  int score = 0;
  ui sq, rank, fyle;
  BB Temp;

  ASSERT(wSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White queen score                                                */
  /********************************************************************/
  Temp = pos->WQBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      rank = row(sq);
      fyle = column(sq);
      ASSERT(rank < 8);
      score += others->wQSqvalue[sq];

      if (wSleepy == 0 || minPieceScore <= middleGameValue)
	{
	  /* Queen in 7th rank */
	  if ((bitCount(pos->WRBoard | pos->WQBoard) > 1) &&
	      bonusFor7thRank(white, rank, fyle, pos))
	    score += BONUS08;
	}

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score > QBonus) QBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore); 
 return score;
}


static int bQBonus(ui bSleepy, int minPieceScore, sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq, rank, fyle;
  BB Temp;

  ASSERT(bSleepy >= 0);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black queen score                                                */
  /********************************************************************/
  Temp = pos->BQBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      rank = row(sq);
      fyle = column(sq);
      ASSERT(rank < 8);
      score -= others->bQSqvalue[sq];

      if (bSleepy == 0 || minPieceScore <= middleGameValue)
	{
	  /* Queen in 7th rank */
	  if ((bitCount(pos->BRBoard | pos->BQBoard) > 1) &&
	      bonusFor7thRank(black, rank, fyle, pos))
	    score -= BONUS08;
	}

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wKBonus(sPos *pos, sOthers *others) 
{
  int score = 0;
  ui sq;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White king safety score                                          */
  /********************************************************************/
  sq = pos->wKSq;
  ASSERT(sq >= A1 && sq <= H8);
  score += others->wKSqvalue[sq];

  if (pos->BQBoard)         /* opponent has queen */
    {
        /* Good neighborhood */
      if (sq == A1 || sq == B1 || sq == G1 || sq == H1)
	{
	  score += bitCount(others->KHood[sq] & 
			    (pos->WBoard & ~(pos->WPBoard))) * BONUS09;
	  ASSERT(score <= mateScore);
	}
    }

#ifdef BONUSES_PENALTIES
  if (score > KBonus) KBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


int wKPenalty(ui ply, sPos *pos, sPlyData *plyData, sOthers *others) 
{
  int score = 0;
  ui sq, sq1, fyle, fylePlus, fyleMinus, rank;
  BB Temp, AttackBoard;
  ui i;
  ui ksfCounter, ksfFlag;

  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);
  ASSERT(plyData != NULL);

  /********************************************************************/
  /* White king safety score                                          */
  /********************************************************************/
  sq = pos->wKSq;
  ASSERT(sq >= A1 && sq <= H8);

  AttackBoard = pos->BQBoard & others->KMoves[sq];
  if (AttackBoard &&             /* opponent's queen is at king's face */
      (getBAttacker(firstOne(AttackBoard), pos, &plyData[ply], others) & 2) &&
      getWAttacker(firstOne(AttackBoard), pos, &plyData[ply], others) == 131)
    {
      score -= ksfTable[39];
    }
  else if (pos->BQBoard)         /* opponent has queen */
    {
      /* Too advanced is dangerous */
      rank = row(sq);
      fyle = column(sq);
      ASSERT(fyle < 8 && rank < 8);
      ksfCounter = 0;

      if (rank < 2)
	{
	  /* pawn cover in front of king */
	  if (!pos->wPCount[fyle])
	    {
	      score -= PENALTY08;
	      if (!pos->bPCount[fyle])
		{
		  score -= PENALTY18;
		}
	    }
	  else if (!(OneMask[sq+8] & pos->WPBoard))
	    {
	      score -= PENALTY06;
	      if (!(OneMask[sq+16] & pos->WPBoard))
		{
		  score -= PENALTY07;
		}
	    }

	  if (fyle > 3)
	    {
	      /* pawn cover outside of king */
	      if (fyle < 7)
		{
		  fylePlus = fyle+1;
		  sq1 = sq+1;
		  if (!pos->wPCount[fylePlus])
		    {
		      score -= PENALTY11;
		      if (!pos->bPCount[fylePlus])
			{
			  score -= PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY09;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY10;
			}
		    }
		}
	      if (fyle < 6)
		{
		  fylePlus = fyle+2;
		  sq1 = sq+2;
		  if (!pos->wPCount[fylePlus])
		    {
		      score -= PENALTY11;
		      if (!pos->bPCount[fylePlus])
			{
			  score -= PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY09;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY10;
			}
		    }
		}
	      if (fyle < 5)
		{
		  fylePlus = fyle+3;
		  sq1 = sq+3;
		  if (!pos->wPCount[fylePlus])
		    {
		      score -= PENALTY11;
		      if (!pos->bPCount[fylePlus])
			{
			  score -= PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY09;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY10;
			}
		    }
		}

	      /* pawn cover inside of king */
	      if (fyle > 4)
		{
		  fyleMinus = fyle-1;
		  sq1 = sq-1;
		  if (!pos->wPCount[fyleMinus])
		    {
		      score -= PENALTY14;
		      if (!pos->bPCount[fyleMinus])
			{
			  score -= PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY12;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY13;
			}
		    }
		}
	      if (fyle > 6)    /* king on G1 doesn't need cover in E file */
		{
		  fyleMinus = fyle-2;
		  sq1 = sq-2;
		  if (!pos->wPCount[fyleMinus])
		    {
		      score -= PENALTY14;
		      if (!pos->bPCount[fyleMinus])
			{
			  score -= PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY12;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY13;
			}
		    }
		}
	      /* basis for "mayor attack" */
	      if ((pos->BPBoard & OneMask[D5]) && 
		  (pos->BPBoard & OneMask[E4])) score -= PENALTY17;
	      if (column(pos->bKSq) < 4) score += score;
	    }
	  else if (fyle < 4)
	    {
	      /* pawn cover outside of king */
	      if (fyle > 0)
		{
		  fyleMinus = fyle-1;
		  sq1 = sq-1;
		  if (!pos->wPCount[fyleMinus])
		    {
		      score -= PENALTY11;
		      if (
			  !pos->bPCount[fyleMinus])
			{
			  score -= PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY09;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY10;
			}
		    }
		}
	      if (fyle > 1)
		{
		  fyleMinus = fyle-2;
		  sq1 = sq-2;
		  if (!pos->wPCount[fyleMinus])
		    {
		      score -= PENALTY11;
		      if (!pos->bPCount[fyleMinus])
			{
			  score -= PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY09;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY10;
			}
		    }
		}
	      if (fyle > 2)
		{
		  fyleMinus = fyle-3;
		  sq1 = sq-3;
		  if (!pos->wPCount[fyleMinus])
		    {
		      score -= PENALTY11;
		      if (!pos->bPCount[fyleMinus])
			{
			  score -= PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY09;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY10;
			}
		    }
		}

	      /* pawn cover inside of king */
	      if (fyle < 3)
		{
		  fylePlus = fyle+1;
		  sq1 = sq+1;
		  if (!pos->wPCount[fylePlus])
		    {
		      score -= PENALTY14;
		      if (!pos->bPCount[fylePlus])
			{
			  score -= PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY12;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY13;
			}
		    }
		}
	      if (fyle < 1)    /* king on B1 doesn't need cover in D file */
		{
		  fylePlus = fyle+2;
		  sq1 = sq+2;
		  if (!pos->wPCount[fylePlus])
		    {
		      score -= PENALTY14;
		      if (!pos->bPCount[fylePlus])
			{
			  score -= PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1+8] & pos->WPBoard))
		    {
		      score -= PENALTY12;
		      if (!(OneMask[sq1+16] & pos->WPBoard))
			{
			  score -= PENALTY13;
			}
		    }
		}
	      /* basis for "mayor attack" */
	      if ((pos->BPBoard & OneMask[E5]) && 
		  (pos->BPBoard & OneMask[D4])) score -= PENALTY17;
	      if (column(pos->bKSq) > 3) score += score;
	    }

	  /* Penalty for B2 or G2 hole */
	  if ((others->KMoves[sq] & OneMask[B2]) && 
	      !((pos->WPBoard | pos->WBBoard) & OneMask[B2])) 
	    score -= PENALTY15;
	  else if ((others->KMoves[sq] & OneMask[G2]) && 
		   !((pos->WPBoard | pos->WBBoard) & OneMask[G2])) 
	    score -= PENALTY15;
	}

      /* Losing castling right */
      if (plyData[0].wCanCastle > 0 && plyData[ply].wCanCastle == 0) 
	score -= PENALTY16;
      
      ksfCounter = -score/32;
      score -= kRankPenalty[rank];

      /* Rebel's evaluation.  See eval.h for explanation */
      ksfCounter += wKsfCounter[sq]; /* initial king safety counter */
      Temp = others->KMoves[sq];  /* squares around the king */
      ksfFlag = 0;
      while (Temp)
	{
	  i = firstOne(Temp);	
	  /* Second bit on means there is an attacker */
	  if (getBAttacker(i, pos, &plyData[ply], others) & 2)
	    {
	      ksfFlag |= plyData[ply].bAttacker[i];
	      ksfCounter++;
	      /* no friendly piece on square in front of king */
	      if (row(i) > rank && (!(OneMask[i] & pos->WBoard)))
		ksfCounter++;
	      /* Only our king defends the square.  131 = 1+2+128 */
	      if (getWAttacker(i, pos, &plyData[ply], others) == 131)
		ksfCounter++;
	    }
	  Temp &= NilMask[i];
	}
      ASSERT((ksfFlag >> 3) < 32);
      ksfCounter += ksfFlagToCounter[ksfFlag >> 3];
      ASSERT(ksfCounter < 60);
      score -= ksfTable[ksfCounter];
    }

#ifdef BONUSES_PENALTIES
  if (score < KPenalty) KPenalty = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bKBonus(sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black king safety score                                          */
  /********************************************************************/
  sq = pos->bKSq;
  ASSERT(sq >= A1 && sq <= H8);
  score -= others->bKSqvalue[sq];

  if (pos->WQBoard)         /* opponent has queen */
    {
      /* Good neighborhood */
      if (sq == A8 || sq == B8 || sq == G8 || sq == H8)
	{
	  score -= bitCount(others->KHood[sq] & 
			    (pos->BBoard & ~(pos->BPBoard))) * BONUS09;
	  ASSERT(score >= -mateScore);
	}
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


int bKPenalty(ui ply, sPos *pos, sPlyData *plyData, sOthers *others)
{
  int score = 0;
  ui sq, sq1, fyle, fylePlus, fyleMinus, rank;
  BB Temp, AttackBoard;
  ui i;
  ui ksfCounter, ksfFlag;

  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);
  ASSERT(plyData != NULL);

  /********************************************************************/
  /* Black king safety score                                          */
  /********************************************************************/
  sq = pos->bKSq;
  ASSERT(sq >= A1 && sq <= H8);

  AttackBoard = pos->WQBoard & others->KMoves[sq];
  if (AttackBoard &&             /* opponent's queen is at king's face */
      (getWAttacker(firstOne(AttackBoard), pos, &plyData[ply], others) & 2) &&
      getBAttacker(firstOne(AttackBoard), pos, &plyData[ply], others) == 131)
    {
      score += ksfTable[39];
    }
  else if (pos->WQBoard)         /* opponent has queen */
    {
      /* Too advanced is dangerous */
      rank = row(sq);
      fyle = column(sq);
      ASSERT(fyle < 8 && rank < 8);
      ksfCounter = 0;
      
      if (rank > 5)
	{
	  /* pawn in front of king */
	  if (!pos->bPCount[fyle])
	    {
	      score += PENALTY08;
	      if (!pos->wPCount[fyle])
		{
		  score += PENALTY18;
		}
	    }
	  else if (!(OneMask[sq-8] & pos->BPBoard))
	    {
	      score +=  PENALTY06;
	      if (!(OneMask[sq-16] & pos->BPBoard))
		{
		  score +=  PENALTY07;
		}
	    }

	  if (fyle > 3)
	    {
	      /* pawn cover outside of king */
	      if (fyle < 7)
		{
		  fylePlus = fyle+1;
		  sq1 = sq+1;
		  if (!pos->bPCount[fylePlus])
		    {
		      score += PENALTY11;
		      if (!pos->wPCount[fylePlus])
			{
			  score += PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY09;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY10;
			}
		    }
		}
	      if (fyle < 6)
		{
		  fylePlus = fyle+2;
		  sq1 = sq+2;
		  if (!pos->bPCount[fylePlus])
		    {
		      score += PENALTY11;
		      if (!pos->wPCount[fylePlus])
			{
			  score += PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY09;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY10;
			}
		    }
		}
	      if (fyle < 5)
		{
		  fylePlus = fyle+3;
		  sq1 = sq+3;
		  if (!pos->bPCount[fylePlus])
		    {
		      score += PENALTY11;
		      if (!pos->wPCount[fylePlus])
			{
			  score += PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY09;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY10;
			}
		    }
		}

	      /* pawn cover inside of king */
	      if (fyle > 4)
		{
		  fyleMinus = fyle-1;
		  sq1 = sq-1;
		  if (!pos->bPCount[fyleMinus])
		    {
		      score += PENALTY14;
		      if (!pos->wPCount[fyleMinus])
			{
			  score += PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY12;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY13;
			}
		    }
		}
	      if (fyle > 6)    /* king on G8 doesn't need cover in E file */
		{
		  fyleMinus = fyle-2;
		  sq1 = sq-2;
		  if (!pos->bPCount[fyleMinus])
		    {
		      score += PENALTY14;
		      if (!pos->wPCount[fyleMinus])
			{
			  score += PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY12;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY13;
			}
		    }
		}
	      /* basis for "mayor attack" */
	      if ((pos->WPBoard & OneMask[D4]) && 
		  (pos->WPBoard & OneMask[E5])) score += PENALTY17;
	      if (column(pos->wKSq) < 4) score += score;
	    }
	  else if (fyle < 4)
	    {
	      /* pawn cover outside of king */
	      if (fyle > 0)
		{
		  fyleMinus = fyle-1;
		  sq1 = sq-1;
		  if (!pos->bPCount[fyleMinus])
		    {
		      score += PENALTY11;
		      if (!pos->wPCount[fyleMinus])
			{
			  score += PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY09;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY10;
			}
		    }
		}
	      if (fyle > 1)
		{
		  fyleMinus = fyle-2;
		  sq1 = sq-2;
		  if (!pos->bPCount[fyleMinus])
		    {
		      score += PENALTY11;
		      if (!pos->wPCount[fyleMinus])
			{
			  score += PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY09;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY10;
			}
		    }
		}
	      if (fyle > 2)
		{
		  fyleMinus = fyle-3;
		  sq1 = sq-3;
		  if (!pos->bPCount[fyleMinus])
		    {
		      score += PENALTY11;
		      if (!pos->wPCount[fyleMinus])
			{
			  score += PENALTY19;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY09;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY10;
			}
		    }
		}

	      /* pawn cover inside of king */
	      if (fyle < 3)
		{
		  fylePlus = fyle+1;
		  sq1 = sq+1;
		  if (!pos->bPCount[fylePlus])
		    {
		      score += PENALTY14;
		      if (!pos->wPCount[fylePlus])
			{
			  score += PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY12;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY13;
			}
		    }
		}
	      if (fyle < 1)    /* king on B1 doesn't need cover in D file */
		{
		  fylePlus = fyle+2;
		  sq1 = sq+2;
		  if (!pos->bPCount[fylePlus])
		    {
		      score += PENALTY14;
		      if (!pos->wPCount[fylePlus])
			{
			  score += PENALTY20;
			}
		    }
		  else if (!(OneMask[sq1-8] & pos->BPBoard))
		    {
		      score += PENALTY12;
		      if (!(OneMask[sq1-16] & pos->BPBoard))
			{
			  score += PENALTY13;
			}
		    }
		}
	      /* basis for "mayor attack" */
	      if ((pos->WPBoard & OneMask[E4]) && 
		  (pos->WPBoard & OneMask[D5])) score += PENALTY17;
	      if (column(pos->wKSq) > 3) score += score;
	    }

	  /* Penalty for B7 or G7 hole */
	  if ((others->KMoves[sq] & OneMask[B7]) && 
	      !((pos->BPBoard | pos->BBBoard) & OneMask[B7])) 
	    score += PENALTY15;
	  else if ((others->KMoves[sq] & OneMask[G7]) && 
		   !((pos->BPBoard | pos->BBBoard) & OneMask[G7])) 
	    score += PENALTY15;
	}

      /* Losing castling right */
      if (plyData[0].bCanCastle > 0 && plyData[ply].bCanCastle == 0) 
	score += PENALTY16;
      
      ksfCounter = score/32;
      score += kRankPenalty[7-rank];

      /* Rebel's evaluation.  See eval.h for explanation */
      ksfCounter += bKsfCounter[sq]; /* initial king safety counter */
      Temp = others->KMoves[sq];  /* squares around the king */
      ksfFlag = 0;
      while (Temp)
	{
	  i = firstOne(Temp);
	  /* Second bit on means there is an attacker */
	  if (getWAttacker(i, pos, &plyData[ply], others) & 2)
	    {
	      ksfFlag |= plyData[ply].wAttacker[i];
	      ksfCounter++;
	      /* no friendly piece on square in front of king */
	      if (row(i) < rank && (!(OneMask[i] & pos->BBoard)))
		ksfCounter++;
	      /* Only our king defends the square.  131 = 1+2+128 */
	      if (getBAttacker(i, pos, &plyData[ply], others) == 131) 
		ksfCounter++;
	    }
	  Temp &= NilMask[i];
	}
      ASSERT((ksfFlag >> 3) < 32);
      ksfCounter += ksfFlagToCounter[ksfFlag >> 3];
      ASSERT(ksfCounter < 60);
      score += ksfTable[ksfCounter];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wPEndBonus(ui wToMove, int bPieceScore, sPos *pos, sOthers *others, 
		      ui bothSidePawns, ui *ppCount)
{
  ui wPassedPSq[8];
  int score = 0;
  ui i, n, sq, ourPawns, theirPawns, pawnDistance, kingDistance;
  ui wDistance, bDistance, minDistance;
  ui rank;

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);
  ASSERT(ppCount != NULL);

  /********************************************************************/
  /* White pawn score                                                 */
  /********************************************************************/

  /* This variable has to be initialized before the real loop. */
  for (i = 0; i < 8; i++) wPassedPSq[i] = 64;
  for (i = 0; i < 8; i++)
    {
      n = pos->wPCount[i];
      if (n)
	{
	  /* doubled or tripled or .... pawns */
	  if (n > 1) score -= (n-1) * PENALTY00;

	  sq = lastOne(pos->WPBoard & others->Column[i]);
	  ASSERT(sq >= A1 && sq <= H8);
	  rank = row(sq);
	  ASSERT(rank < 8);
	  wDistance = (ui) distance(sq, pos->wKSq);
	  bDistance = (ui) distance(sq, pos->bKSq);
	  ASSERT(wDistance < 8 && bDistance < 8);
	  if (wDistance < bDistance) minDistance = wDistance;
	  else minDistance = bDistance;

	  /* Any pawn likes black king, hates white king. */
	  score += rank * (bDistance - wDistance);
	  if (bPieceScore == 0)
	    {
	      score += KING_RACE_FACTOR * (bDistance - wDistance) *
		(7-minDistance);
	    }
	  /* passed pawns */
	  if (!(pos->BPBoard & others->Up8Of[sq]))
	    {
	      /* Passed pawn likes white king more, hates black king more. */
	      score += rank * (bDistance - wDistance);
	      if (bPieceScore == 0)
		{
		  score += KING_RACE_FACTOR * (bDistance - wDistance) *
		    (7-minDistance);
		}

	      theirPawns = 
		bitCount(pos->BPBoard & others->Up8SideOf[sq]);
	      ASSERT(theirPawns <= 8);
	      if (!theirPawns)
		{
		  *ppCount = *ppCount + 1;
		  wPassedPSq[i] = sq;
		  score += others->wPassedPSqvalue[sq] + rank * 
		    (bDistance - wDistance);
		  /* protected */
		  if (others->BPCaptures[sq] & pos->WPBoard) 
		    score += others->wPassedPSqvalue[sq] >> 1;
	      
		  /* pawn king square rule */
		  if (bPieceScore == 0)
		    {
		      pawnDistance = (ui) distance(sq, wPGoal[i]);
		      kingDistance = (ui) distance(pos->bKSq, wPGoal[i]);
		      ASSERT(pawnDistance < 8 && kingDistance < 8);
		      if (pawnDistance+1 < kingDistance) 
			{
			  score += rank * pValue;
			}
		      else if (wToMove && pawnDistance < kingDistance)
			{
			  score += rank * pValue;
			}
		    }

		  /* pair of passed pawns */
		  ASSERT(i >= 0 && i <= 7);
		  ASSERT(sq >= A1 && sq <= H8);
		  if (i > 0 && (sq-1 == wPassedPSq[i-1])) 
		    {
		      /* *ppCount = *ppCount + 1; */
		      score += pppBonus[rank];
		    }

		  /* rook behind passed pawn */
		  if (pos->wPCount[i] == 1)
		    {
		      if (pos->WRBoard & others->Down8Of[sq]) 
			score += BONUS10;
		      if (pos->BRBoard & others->Down8Of[sq]) 
			score -= BONUS10;
		    }
		}
	      else
		{
		  ourPawns = bitCount(pos->WPBoard & 
			       (others->SideOf[sq] | others->Down8SideOf[sq]));
		  ASSERT(ourPawns <= 8);
		  if (ourPawns >= theirPawns) 
		    {
		      score += others->wPassedPSqvalue[sq] + rank * 
			(bDistance - wDistance);

		      /* protected */
		      if (others->BPCaptures[sq] & pos->WPBoard) 
			score += others->wPassedPSqvalue[sq] >> 1;
	      
		      /* rook behind passed pawn */
		      if (pos->wPCount[i] == 1)
			{
			  if (pos->WRBoard & others->Down8Of[sq]) 
			    score += BONUS10;
			  if (pos->BRBoard & others->Down8Of[sq]) 
			    score -= BONUS10;
			}
		    }
		}
	    }

	  /* strength in company */
	  if (others->SideOf[sq] & pos->WPBoard) score += rank*BONUS11;
	}
    }

  /* With pawns at both sides of the board, bishop is better than knight */
  if (bothSidePawns && pos->WBBoard && 
      bPieceScore > -rValue && !pos->BBBoard) 
    score += BONUS12;

#ifdef BONUSES_PENALTIES
  if (score > PEndBonus) PEndBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bPEndBonus(ui wToMove, int wPieceScore, sPos *pos, sOthers *others, 
		      ui bothSidePawns, ui *ppCount)
{
  ui bPassedPSq[8];
  int score = 0;
  ui i, n, sq, ourPawns, theirPawns, pawnDistance, kingDistance;
  ui wDistance, bDistance, minDistance;
  ui rank;

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);
  ASSERT(ppCount != NULL);

  /********************************************************************/
  /* Black pawn score                                                 */
  /********************************************************************/

  /* This variable has to be initialized before the real loop. */
  for (i = 0; i < 8; i++) bPassedPSq[i] = 64;
  for (i = 0; i < 8; i++)
    {
      n = pos->bPCount[i];
      if (n)
	{
	  /* doubled or tripled or .... pawns */
	  if (n > 1) score += (n-1) * PENALTY00;

	  sq = firstOne(pos->BPBoard & others->Column[i]);
	  ASSERT(sq >= A1 && sq <= H8);
	  rank = 7-row(sq);
	  ASSERT(rank < 8);
	  wDistance = (ui) distance(sq, pos->wKSq);
	  bDistance = (ui) distance(sq, pos->bKSq);
	  ASSERT(wDistance < 8 && bDistance < 8);
	  if (wDistance < bDistance) minDistance = wDistance;
	  else minDistance = bDistance;

	  /* Any pawn likes black king, hates white king. */
	  score -= rank * (wDistance - bDistance);
	  if (wPieceScore == 0)
	    {
	      score -= KING_RACE_FACTOR * (wDistance - bDistance) *
		(7-minDistance);
	    }
	  /* passed pawns */
	  if (!(pos->WPBoard & others->Down8Of[sq]))
	    {
	      /* Passed pawn likes black king more, hates white king more. */
	      score -= rank * (wDistance - bDistance);
	      if (wPieceScore == 0)
		{
		  score -= KING_RACE_FACTOR * (wDistance - bDistance) *
		    (7-minDistance);
		}

	      theirPawns = 
		bitCount(pos->WPBoard & others->Down8SideOf[sq]);
	      ASSERT(theirPawns <= 8);
	      if (!theirPawns)
		{
		  *ppCount = *ppCount + 1;
		  bPassedPSq[i] = sq;
		  score -= (others->bPassedPSqvalue[sq] + rank * 
			    (wDistance - bDistance));

		  /* protected */
		  if (others->WPCaptures[sq] & pos->BPBoard) 
		    score -= others->bPassedPSqvalue[sq] >> 1;
		  
		  /* pawn king square rule */
		  if (wPieceScore == 0)
		    {
		      pawnDistance = (ui) distance(sq, bPGoal[i]);
		      kingDistance = (ui) distance(pos->wKSq, bPGoal[i]);
		      ASSERT(pawnDistance < 8 && kingDistance < 8);
		      if (pawnDistance+1 < kingDistance) 
			{
			  score -= rank * pValue;
			}
		      else if (!wToMove && pawnDistance < kingDistance)
			{
			  score -= rank * pValue;
			}
		    }

		  /* pair of passed pawns */
		  ASSERT(i >= 0 && i <= 7);
		  ASSERT(sq >= A1 && sq <= H8);
		  if (i > 0 && (sq-1 == bPassedPSq[i-1])) 
		    {
		      /* *ppCount = *ppCount + 1; */
		      score -= pppBonus[rank];
		    }

		  /* rook behind passed pawn */
		  if (pos->bPCount[i] == 1)
		    {
		      if (pos->WRBoard & others->Up8Of[sq]) 
			score += BONUS10;
		      if (pos->BRBoard & others->Up8Of[sq]) 
			score -= BONUS10;
		    }
		}
	      else
		{
		  ourPawns = bitCount(pos->BPBoard & 
			       (others->SideOf[sq] | others->Up8SideOf[sq]));
		  ASSERT(ourPawns <= 8);
		  if (ourPawns >= theirPawns) 
		    {
		      score -= (others->bPassedPSqvalue[sq] + rank * 
				(wDistance - bDistance));

		      /* protected */
		      if (others->WPCaptures[sq] & pos->BPBoard) 
			score -= others->bPassedPSqvalue[sq] >> 1;
	      
		      /* rook behind passed pawn */
		      if (pos->bPCount[i] == 1)
			{
			  if (pos->WRBoard & others->Up8Of[sq]) 
			    score += BONUS10;
			  if (pos->BRBoard & others->Up8Of[sq]) 
			    score -= BONUS10;
			}
		    }
		}
	    }

	  /* strength in company */
	  if (others->SideOf[sq] & pos->BPBoard) score -= rank*BONUS11;
	}
    }

  /* With pawns at both sides of the board, bishop is better than knight */
  if (bothSidePawns && pos->BBBoard &&
      wPieceScore < rValue && !pos->WBBoard) 
    score -= BONUS12;

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wNEndBonus(sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White knight score                                                */
  /********************************************************************/

  if (pos->WPBoard)
    {
      Temp = pos->WNBoard;
      while (Temp)
	{
	  sq = firstOne(Temp);
	  ASSERT(sq >= A1 && sq <= H8);
	  score += others->wNSqvalue[sq];
	  AttackBoard = others->NMoves[sq] & ~pos->WBoard;
	  score += bitCount(AttackBoard) * BONUS00;
	  ASSERT(score <= mateScore);
	  Temp &= NilMask[sq];
	}
    }

#ifdef BONUSES_PENALTIES
  if (score > NEndBonus) NEndBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bNEndBonus(sPos *pos, sOthers *others)
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black knight score                                               */
  /********************************************************************/

  if (pos->BPBoard)
    {
      Temp = pos->BNBoard;
      while (Temp)
	{
	  sq = firstOne(Temp);
	  ASSERT(sq >= A1 && sq <= H8);
	  score -= others->bNSqvalue[sq];
	  AttackBoard = others->NMoves[sq] & ~pos->BBoard;
	  score -= bitCount(AttackBoard) * BONUS00;
	  ASSERT(score >= -mateScore);
	  Temp &= NilMask[sq];
	}
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wBEndBonus(sPos *pos)
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* White bishop score                                               */
  /********************************************************************/

  if (pos->WPBoard)
    {
      Temp = pos->WBBoard;
      while (Temp)
	{
	  sq = firstOne(Temp);
	  ASSERT(sq >= A1 && sq <= H8);
	  AttackBoard = bishopAttacks(sq, pos) & ~pos->WBoard;
	  score += bitCount(AttackBoard) * BONUS05;
	  ASSERT(score <= mateScore);
	  Temp &= NilMask[sq];
	}
    }

#ifdef BONUSES_PENALTIES
  if (score > BEndBonus) BEndBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bBEndBonus(sPos *pos)
{
  int score = 0;
  ui sq;
  BB Temp, AttackBoard;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* Black bishop score                                                */
  /********************************************************************/
  if (pos->BPBoard)
    {
      Temp = pos->BBBoard;
      while (Temp)
	{
	  sq = firstOne(Temp);
	  ASSERT(sq >= A1 && sq <= H8);
	  AttackBoard = bishopAttacks(sq, pos) & ~pos->BBoard;
	  score -= bitCount(AttackBoard) * BONUS05;
	  ASSERT(score >= -mateScore);
	  Temp &= NilMask[sq];
	}
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wREndBonus(sPos *pos)
{
  int score = 0;
  ui sq, rank, fyle;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* White rook score                                                 */
  /********************************************************************/
  Temp = pos->WRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      rank = row(sq);
      fyle = column(sq);
      ASSERT(fyle < 8);

      /* semi open or open file */
      if (!pos->wPCount[fyle])
	{
	  if (!pos->bPCount[fyle]) score += BONUS15;
	}

      /* Rook in 7th rank */
      if (bonusFor7thRank(white, rank, fyle, pos)) score += BONUS07;

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score > REndBonus) REndBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wREndPenalty(sPos *pos)
{
  int score = 0;
  ui sq, fyle;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* White rook score                                                 */
  /********************************************************************/
  Temp = pos->WRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      fyle = column(sq);
      ASSERT(fyle < 8);

      /* Rook obstructed by pawn */
      if (fyle == 0 && 
	  ((pos->WPBoard & OneMask[sq+1]) ||
	   (pos->WPBoard & OneMask[sq+2])))
	score -= PENALTY04;
      else if (fyle == 7 && 
	       ((pos->WPBoard & OneMask[sq-1]) ||
		(pos->WPBoard & OneMask[sq-2])))
	score -= PENALTY04;

      Temp &= NilMask[sq];
    }

#ifdef BONUSES_PENALTIES
  if (score < REndPenalty) REndPenalty = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bREndBonus(sPos *pos)
{
  int score = 0;
  ui sq, rank, fyle;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* Black rook score                                                 */
  /********************************************************************/
  Temp = pos->BRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      rank = row(sq);
      fyle = column(sq);
      ASSERT(fyle < 8);

      /* semi open or open file */
      if (!pos->bPCount[fyle])
	{
	  if (!pos->wPCount[fyle]) score -= BONUS15;
	}

      /* Rook in 7th rank */
      if (bonusFor7thRank(black, rank, fyle, pos)) score -= BONUS07;

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bREndPenalty(sPos *pos)
{
  int score = 0;
  ui sq, fyle;
  BB Temp;

  ASSERT(pos != NULL);

  /********************************************************************/
  /* Black rook score                                                 */
  /********************************************************************/
  Temp = pos->BRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      ASSERT(sq >= A1 && sq <= H8);
      fyle = column(sq);
      ASSERT(fyle < 8);

      /* Rook obstructed by pawn */
      if (fyle == 0 && 
	  ((pos->BPBoard & OneMask[sq+1]) ||
	   (pos->BPBoard & OneMask[sq+2])))
	score += PENALTY04;
      else if (fyle == 7 && 
	       ((pos->BPBoard & OneMask[sq-1]) ||
		(pos->BPBoard & OneMask[sq-2])))
	score += PENALTY04;

      Temp &= NilMask[sq];
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int wKEndBonus(int wPieceScore, int bPieceScore, sPos *pos,
		      sOthers *others)
{
  int score = 0;
  ui fyle, rFile;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* White king score */
  /********************************************************************/

  /* if we can mate, then push opponent king */
  if (wPieceScore+bPieceScore >= rValue)
    score += (7-distance(pos->wKSq, pos->bKSq)) * BONUS13;

  /* Good neighborhood */
  if (pos->BQBoard  || bitCount(pos->BRBoard) == 2)
    {
      score += bitCount(others->KHood[pos->wKSq] & pos->WBoard) * BONUS09;
      ASSERT(score <= mateScore);
    }

  /* Shelter the king from that pesky rook */
  if (bitCount(pos->BRBoard) == 1)
    {
      rFile = column(firstOne(pos->BRBoard));
      fyle = column(pos->wKSq);
      if ((fyle+1 < rFile) && (pos->WPBoard & OneMask[pos->wKSq+1]))
	score += BONUS14;
      else if ((fyle > rFile+1) && (pos->WPBoard & OneMask[pos->wKSq-1]))
	score += BONUS14;
    }

#ifdef BONUSES_PENALTIES
  if (score > KEndBonus) KEndBonus = score;
#endif

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


static int bKEndBonus(int wPieceScore, int bPieceScore, sPos *pos,
		      sOthers *others)
{
  int score = 0;
  ui fyle, rFile;

  ASSERT(pos != NULL);
  ASSERT(others != NULL);

  /********************************************************************/
  /* Black king score */
  /********************************************************************/

  /* if we can mate, then push opponent king */
  if (wPieceScore+bPieceScore <= -rValue)
    score = (distance(pos->wKSq, pos->bKSq)-7) * BONUS13;

  /* Good neighborhood */
  if (pos->WQBoard  || bitCount(pos->WRBoard) == 2)
    {
      score -= bitCount(others->KHood[pos->bKSq] & pos->BBoard) * BONUS09;
      ASSERT(score >= -mateScore);
    }

  /* Shelter the king from that pesky rook */
  if (bitCount(pos->WRBoard) == 1)
    {
      rFile = column(firstOne(pos->WRBoard));
      fyle = column(pos->bKSq);
      if ((fyle+1 < rFile) && (pos->BPBoard & OneMask[pos->bKSq+1]))
	score -= BONUS14;
      else if ((fyle > rFile+1) && (pos->BPBoard & OneMask[pos->bKSq-1]))
	score -= BONUS14;
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


/* Is this position a draw by rule? */
ui isDrawByRule(ui ply, sPos *pos, sPlyData *thisPlyData) 
{
  ui wCantWin = false, bCantWin = false;
  int i, stillGoing;
  ui happenings = 1;

  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(pos != NULL);
  ASSERT(thisPlyData != NULL);

  /* Not enough material to win */
  if (bitCount(pos->WPBoard) == 0 && bitCount(pos->BPBoard) == 0)
    {
      if (bitCount(pos->WBoard) <= 3 && 
	  !bitCount(pos->WQBoard) && 
	  !bitCount(pos->WRBoard))
	{
	  switch(bitCount(pos->WBBoard))
	    {
	    case 0:
	      wCantWin = true;
	      break;
	    case 1:
	      if (!bitCount(pos->WNBoard)) wCantWin = true;
	      break;
	    default:
	      break;
	    }
	}

      if (bitCount(pos->BBoard) <= 3 && 
	  !bitCount(pos->BQBoard) && 
	  !bitCount(pos->BRBoard))
	{
	  switch(bitCount(pos->BBBoard))
	    {
	    case 0:
	      bCantWin = true;
	      break;
	    case 1:
	      if (!bitCount(pos->BNBoard)) bCantWin = true;
	      break;
	    default:
	      break;
	    }
	}

      if (wCantWin && bCantWin) return 13;
    }

  if (thisPlyData->count50Moves > 99) return 50;
  else
    {
      /* According to the rule, a game is drawn if a position
	 happens 3 times, not repeated 3 times (happens 4 times).
      */
      stillGoing = (int) thisPlyData->count50Moves;
      ASSERT(pos->repHead < maxMove);
      for (i = pos->repHead+ply-3; 
	   i >= 0 && stillGoing >= 0; 
	   i -= 2, stillGoing -= 2)
	if (pos->repKey[i] == thisPlyData->key) happenings++;

      if (happenings > 1) return happenings;

      return 0;
    }
}


/* IMPORTANT: evalMaterial() returns white's score */
int eval(ui wToMove, ui ply, sPos *pos, sPlyData *plyData, sOthers *others,
	 int lScore, int wPieceScore, int bPieceScore, int alpha, 
	 int doubleBonus)
{
  ui i, wSleepy, bSleepy, openness = 0, bothSidePawns = 0, ppCount;
  int minPieceScore, score = 0, tempScore = 0, dbReduction;

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(ply >= 0 && ply < maxPly);
  ASSERT(pos != NULL);
  ASSERT(plyData != NULL);
  ASSERT(others != NULL);

  score = lScore;

  if ((wPieceScore + bPieceScore) >= 0) minPieceScore = -bPieceScore;
  else minPieceScore = wPieceScore;

  if (minPieceScore <= endGameValue)
    {
      if (((pos->WPBoard | pos->BPBoard) & others->FarQueenSide) &&
	  ((pos->WPBoard | pos->BPBoard) & others->FarKingSide))
	bothSidePawns = 1;

      if (wToMove)
	{ 
	  /* credit */
	  ppCount = 0;
	  tempScore += bPEndBonus(wToMove, wPieceScore, pos, 
				  others, bothSidePawns, &ppCount);
	  if (ppCount > 0 && doubleBonus > 0)
	    {
	      dbReduction = min(ppCount*PENALTY21, doubleBonus);
	      tempScore -= dbReduction;
	    }
	  tempScore += bBEndBonus(pos);
	  tempScore += wREndPenalty(pos);
	  tempScore += bNEndBonus(pos, others);
	  tempScore += bKEndBonus(wPieceScore, bPieceScore, pos, others);
	  tempScore += bREndBonus(pos);

	  /* debit */
	  ppCount = 0;
	  tempScore += wPEndBonus(wToMove, bPieceScore, pos, 
				  others, bothSidePawns, &ppCount);  /* 3523 */
	  if (ppCount > 0 && doubleBonus < 0)
	    {
	      dbReduction = max(-ppCount*PENALTY21, doubleBonus);
	      tempScore -= dbReduction;
	    }
	  tempScore += wREndBonus(pos);   /* 130 */
	  tempScore += wBEndBonus(pos);   /* 108 */
	  tempScore += wKEndBonus(wPieceScore, bPieceScore, pos, others);   /* 100 */
	  tempScore += wNEndBonus(pos, others);   /* 62 */
	  tempScore += bREndPenalty(pos);   /* 50 */
	}
      else
	{
	  /* credit */
	  ppCount = 0;
	  tempScore += wPEndBonus(wToMove, bPieceScore, pos, 
				  others, bothSidePawns, &ppCount);
	  if (ppCount > 0 && doubleBonus < 0)
	    {
	      dbReduction = max(-ppCount*PENALTY21, doubleBonus);
	      tempScore -= dbReduction;
	    }
	  tempScore += wBEndBonus(pos);
	  tempScore += bREndPenalty(pos);
	  tempScore += wREndBonus(pos);
	  tempScore += wNEndBonus(pos, others);
	  tempScore += wKEndBonus(wPieceScore, bPieceScore, pos, others);

	  /* debit */
	  ppCount = 0;
	  tempScore += bPEndBonus(wToMove, wPieceScore, pos, 
				  others, bothSidePawns, &ppCount);  /* 3523 */
	  if (ppCount > 0 && doubleBonus > 0)
	    {
	      dbReduction = min(ppCount*PENALTY21, doubleBonus);
	      tempScore -= dbReduction;
	    }
	  tempScore += bREndBonus(pos);   /* 130 */
	  tempScore += bBEndBonus(pos);   /* 108 */
	  tempScore += bKEndBonus(wPieceScore, bPieceScore, pos, others); /* 100 */
	  tempScore += bNEndBonus(pos, others);   /* 62 */
	  tempScore += wREndPenalty(pos);   /* 50 */
	}
      ASSERT(tempScore >= -mateScore && tempScore <= mateScore);

      if ((pos->Flags & OneMask[8]) ||        /* drawish */
	  /* queens tend to make end game drawish */
	  (tempScore > 0 && pos->BQBoard) ||
	  (tempScore < 0 && pos->WQBoard))
	tempScore = tempScore/4;
      score += tempScore;
    }
  else
    {
      /* sleepy means not awake/developed */
      wSleepy = bSleepy = 0;
      wSleepy += bitCount(pos->WNBoard & others->Row[0]);
      wSleepy += bitCount(pos->WBBoard & others->Row[0]);
      wSleepy += bitCount(pos->WRBoard & (OneMask[A1] | OneMask[H1]));
      bSleepy += bitCount(pos->BNBoard & others->Row[7]);
      bSleepy += bitCount(pos->BBBoard & others->Row[7]);
      bSleepy += bitCount(pos->BRBoard & (OneMask[A8] | OneMask[H8]));
      ASSERT(wSleepy <= 8 && bSleepy <= 8);
      if (wSleepy != bSleepy)
	{
	  for (i = 0; i < 8; i++)
	    if (!pos->wPCount[i] && !pos->bPCount[i]) openness++;
	  if (bSleepy > wSleepy) score += openness * BONUS16;
	  else if (bSleepy < wSleepy) score -= openness * BONUS16;
	}
      
      /* use "do {} while (false)" loop to allow breaking out */
      do
	{
	  if (wToMove)
	    { 
	      /* credit */
	      score += wKPenalty(ply, pos, plyData, others);
	      score += bPBonus(bSleepy, pos, others);
	      score += bNBonus(bSleepy, pos, others);
	      score += bRBonus(bSleepy, minPieceScore, pos, others);
	      score += bBBonus(pos, others);
	      score += wRPenalty(pos);
	      score += bKBonus(pos, others);
	      score += wBPenalty(pos);
	      score += bQBonus(wSleepy, minPieceScore, pos, others);
	      
	      /* debit */
	      score += bKPenalty(ply, pos, plyData, others);   /* 600 */
	      if (score + 1133 <= alpha) break;

	      score += wPBonus(wSleepy, pos, others); /* 402 */
	      if (score + 731 <= alpha) break;

	      score += wNBonus(wSleepy, pos, others);   /* 344 */
	      if (score + 387 <= alpha) break;

	      score += wRBonus(wSleepy, minPieceScore, pos, others); /* 152 */
	      if (score + 235 <= alpha) break;

	      score += wBBonus(pos, others);   /* 94 */
	      if (score + 141 <= alpha) break;

	      score += bRPenalty(pos);   /* 75 */
	      score += bBPenalty(pos);   /* 30 */
	      score += wKBonus(pos, others);   /* 24 */
	      score += wQBonus(wSleepy, minPieceScore, pos, others);   /* 12 */
	    }
	  else
	    { 
	      /* credit */
	      score += bKPenalty(ply, pos, plyData, others);
	      score += wPBonus(wSleepy, pos, others);
	      score += wNBonus(wSleepy, pos, others);
	      score += wRBonus(wSleepy, minPieceScore, pos, others);
	      score += wBBonus(pos, others);
	      score += bRPenalty(pos);
	      score += wKBonus(pos, others);
	      score += bBPenalty(pos);
	      score += wQBonus(wSleepy, minPieceScore, pos, others);
	      
	      /* debit */
	      score += wKPenalty(ply, pos, plyData, others);   /*600 */
	      if (-score + 1133 <= alpha) break;

	      score += bPBonus(bSleepy, pos, others); /* 402 */
	      if (-score + 731 <= alpha) break;

	      score += bNBonus(bSleepy, pos, others);   /* 344 */
	      if (-score + 387 <= alpha) break;

	      score += bRBonus(bSleepy, minPieceScore, pos, others); /* 152 */
	      if (-score + 235 <= alpha) break;

	      score += bBBonus(pos, others);   /* 94 */
	      if (-score + 141 <= alpha) break;

	      score += wRPenalty(pos);   /* 75 */
	      score += wBPenalty(pos);   /* 30 */
	      score += bKBonus(pos, others);   /* 24 */
	      score += bQBonus(wSleepy, minPieceScore, pos, others);   /* 12 */
	    }
	} while (false);
    }

  ASSERT(score >= -mateScore && score <= mateScore);
  return score;
}


/********************** Root evaluation ***********************/


void revalWOpening(BB move, sPos *pos) 
{
  ASSERT(from(move) <= H8 && to(move) <= H8 &&
	 from(move) != to(move));
  ASSERT(piece(move) < 13);
  ASSERT(pos != NULL);
  ASSERT(pos->rOpening < 4 && pos->moveNumber < maxPly);

  /********************************************************************/
  /* If out of opening early, use Purdy's All Purpose System (Colle)  */
  /********************************************************************/

  if (pos->rOpening == 1 && pos->moveNumber < 7)
    {
      /* d4 */
      if (piece(move) == wP && from(move) == D2 && to(move) == D4)
	pos->wScore += 32;

      /* Nf3 */
      else if (piece(move) == wN && from(move) == G1 && to(move) == F3)
	pos->wScore += 30;

      /* e3 */
      else if (piece(move) == wP && from(move) == E2 && to(move) == E3)
	pos->wScore += 28;

      /* Bd3 */
      else if (!(pos->WPBoard & OneMask[D2]) &&        /* no pawn on D2 */
	       piece(move) == wB && from(move) == F1 && to(move) == D3)
	pos->wScore += 26;

      /* Nbd2 */
      else if (piece(move) == wN && from(move) == B1 && to(move) == D2)
	pos->wScore += 24;

      /* O-O */
      else if (piece(move) == wK && from(move) == E1 && to(move) == G1)
	pos->wScore += 22;

      /* c3 */
      else if (piece(move) == wP && from(move) == C2 && to(move) == C3)
	pos->wScore += 20;
    }


  /********************************************************************/
  /* If out of opening early, use King's Indian Attack (Pirc flipped) */
  /* 1. Nf3 2. g3 3. Bg2 4. O-O 5. d3 6. Nc3 7. e4                    */
  /********************************************************************/

  if (pos->rOpening == 2 && pos->moveNumber < 7)
    {
      /* Nf3 */
      if (piece(move) == wN && from(move) == G1 && to(move) == F3)
	pos->wScore += 32;

      /* g3 */
      else if (piece(move) == wP && from(move) == G2 && to(move) == G3)
	pos->wScore += 30;

      /* Bg2 */
      else if (piece(move) == wB && from(move) == F1 && to(move) == G2)
	pos->wScore += 28;

      /* O-O */
      else if (piece(move) == wK && from(move) == E1 && to(move) == G1)
	pos->wScore += 26;

      /* d3 */
      else if (piece(move) == wP && from(move) == D2 && to(move) == D3)
	pos->wScore += 24;

      /* Nc3 */
      else if (piece(move) == wN && from(move) == B1 && to(move) == C3)
	pos->wScore += 22;

      /* e4 */
      else if (piece(move) == wP && from(move) == E2 && to(move) == E4)
	pos->wScore += 20;
    }


  /********************************************************************/
  /* If out of opening early, use Richter-Verosov                     */
  /* 1. d4 2. Nc3 3. Bg5 4. Bxf6 5. e3 6. Bd3 7. Qf3                  */
  /********************************************************************/

  if (pos->rOpening == 3 && pos->moveNumber < 7)
    {
      /* d4 */
      if (piece(move) == wP && from(move) == D2 && to(move) == D4)
	pos->wScore += 32;

      /* Nc3 */
      else if (piece(move) == wN && from(move) == B1 && to(move) == C3)
	pos->wScore += 30;

      /* Bg5 */
      else if (piece(move) == wB && from(move) == C1 && to(move) == G5)
	pos->wScore += 28;

      /* Bxf6 */
      else if ((pos->BNBoard & OneMask[F6]) &&        /* capture bN */
	       piece(move) == wB && from(move) == G5 && to(move) == F6)
	pos->wScore += 26;

      /* e3 */
      else if (piece(move) == wP && from(move) == E2 && to(move) == E3)
	pos->wScore += 24;

      /* Bd3 */
      else if (!(pos->WPBoard & OneMask[D2]) &&        /* no pawn on D2 */
	       piece(move) == wB && from(move) == F1 && to(move) == D3)
	pos->wScore += 26;

      /* Qf3 */
      else if (piece(move) == wQ && from(move) == D1 && to(move) == F3)
	pos->wScore += 20;
    }
}


/* Root evaluation at the root */
void revalBOpening(BB move, sPos *pos) 
{
  ASSERT(from(move) <= H8 && to(move) <= H8 &&
	 from(move) != to(move));
  ASSERT(piece(move) < 13);
  ASSERT(pos != NULL);
  ASSERT(pos->rOpening < 4 && pos->moveNumber < maxPly);

  /********************************************************************/
  /* If out of opening early, use Purdy's All Purpose System          */
  /********************************************************************/

  if (pos->rOpening == 1 && pos->moveNumber < 7)
    {
      /* d5 */
      if (piece(move) == bP && from(move) == D7 && to(move) == D5)
	pos->wScore -= 32;

      /* e6 */
      else if (piece(move) == bP && from(move) == E7 && to(move) == E6)
	pos->wScore -= 30;

      /* Nf6 */
      else if (piece(move) == bN && from(move) == G8 && to(move) == F6)
	pos->wScore -= 28;

      /* Be7 */
      else if (piece(move) == bB && from(move) == F8 && to(move) == E7)
	pos->wScore -= 26;

      /* O-O */
      else if (piece(move) == bK && from(move) == E8 && to(move) == G8)
	pos->wScore -= 24;

      /* b6 */
      else if (piece(move) == bP && from(move) == B7 && to(move) == B6)
	pos->wScore -= 22;
      
      /* c5 */
      else if (piece(move) == bP && from(move) == C7 && to(move) == C5)
	pos->wScore -= 20;
    }


  /*********************************************************************/
  /* If out of opening early, use Richter-Veresov                      */
  /* 1. .. f5 2. .. Nf6 3. .. e6 4. .. Be7 5. .. O-O 5. .. d5 7. .. c6 */
  /*********************************************************************/

  if (pos->rOpening == 2 && pos->moveNumber < 7)
    {
      /* f5 */
      if (piece(move) == bP && from(move) == F7 && to(move) == F5)
	pos->wScore -= 32;

      /* Nf6 */
      else if (piece(move) == bN && from(move) == G8 && to(move) == F6)
	pos->wScore -= 30;

      /* e6 */
      else if (piece(move) == bP && from(move) == E7 && to(move) == E6)
	pos->wScore -= 28;

      /* Be7 */
      else if (piece(move) == bB && from(move) == F8 && to(move) == E7)
	pos->wScore -= 26;

      /* O-O */
      else if (piece(move) == bK && from(move) == E8 && to(move) == G8)
	pos->wScore -= 24;

      /* d5 */
      else if (piece(move) == bP && from(move) == D7 && to(move) == D5)
	pos->wScore -= 22;

      /* c6 */
      else if (piece(move) == bP && from(move) == C7 && to(move) == C6)
	pos->wScore -= 20;
    }


  /**********************************************************************/
  /* If out of opening early, use Pirc (King's Indian Attack flipped)   */
  /* 1. .. d6 2. .. Nf6 3. .. g6 4. .. Bg7 5. .. Nc6 6. .. O-O 7. .. e5 */
  /**********************************************************************/

  if (pos->rOpening == 3 && pos->moveNumber < 7)
    {
      /* d6 */
      if (piece(move) == bP && from(move) == D7 && to(move) == D6)
	pos->wScore -= 32;
      
      /* Nf6 */
      else if (piece(move) == bN && from(move) == G8 && to(move) == F6)
	pos->wScore -= 30;

      /* g6 */
      else if (piece(move) == bP && from(move) == G7 && to(move) == G6)
	pos->wScore -= 28;

      /* Bg7 */
      else if (piece(move) == bB && from(move) == F8 && to(move) == G7)
	pos->wScore -= 26;

      /* Nc6 */
      else if (piece(move) == bN && from(move) == B8 && to(move) == C6)
	pos->wScore -= 24;

      /* O-O */
      else if (piece(move) == bK && from(move) == E8 && to(move) == G8)
	pos->wScore -= 22;
      
      /* e5 */
      else if (piece(move) == bP && from(move) == E7 && to(move) == E5)
	pos->wScore -= 20;
    }
}


/* Root evaluation.  Calculate white's score and store it in pos->wScore. */
void reval(BB move, sPos *pos, sOthers *others)
{
  ui wSleepy, bSleepy;

  ASSERT(from(move) <= H8 && to(move) <= H8 &&
	 from(move) != to(move));
  ASSERT(piece(move) < 13);
  ASSERT(pos != NULL);
  ASSERT(others != NULL);
  
  /* sleepy means not awake/developed */
  wSleepy = bSleepy = 0;
  wSleepy += bitCount(pos->WNBoard & others->Row[0]);
  wSleepy += bitCount(pos->WBBoard & others->Row[0]);
  wSleepy += bitCount(pos->WRBoard & (OneMask[A1] | OneMask[H1]));
  bSleepy += bitCount(pos->BNBoard & others->Row[7]);
  bSleepy += bitCount(pos->BBBoard & others->Row[7]);
  bSleepy += bitCount(pos->BRBoard & (OneMask[A8] | OneMask[H8]));
  ASSERT(wSleepy <= 8 && bSleepy <= 8);
      
  /* White is not developed yet */
  if (wSleepy > 0) revalWOpening(move, pos);

  /* Black is not developed yet */
  if (bSleepy > 0) revalBOpening(move, pos);
}
