/* util.c */
/* Last modified: July 24, 2007 */


static ui reverseSq[] = {56, 57, 58, 59, 60, 61, 62, 63,
			 48, 49, 50, 51, 52, 53, 54, 55,
			 40, 41, 42, 43, 44, 45, 46, 47,
			 32, 33, 34, 35, 36, 37, 38, 39,
			 24, 25, 26, 27, 28, 29, 30, 31,
			 16, 17, 18, 19, 20, 21, 22, 23,
			  8,  9, 10, 11, 12, 13, 14, 15,
			  0,  1,  2,  3,  4,  5,  6,  7};
static uc fOne[65536], lOne[65536];



/* "borrowed" from Crafty */
void initFLOne()
{
  int i, j;
  int maskl, maskr;

  fOne[0] = 16;
  lOne[0] = 16;
  for (i = 1; i < 65536; i++)
    {
      maskl = 32768;
      for (j = 0; j < 16; j++)
	{
	  if ((maskl & i)) 
	    {
	      lOne[i] = j;
	      break;
	    }
	  maskl = maskl >> 1;
	}
      maskr = 1;
      for (j = 0; j < 16; j++)
	{
	  if ((maskr & i))
	    {
	      fOne[i] = j;
	      break;
	    }
	  maskr = maskr << 1;
	}
    }
}


/* Translate move into SAN notation */
void san(ui wToMove, ui ply, BB move, char *s, sPos *pos, sPlyData *plyData,
	 sOthers *others)
{
  ui sq, t, p, pT;
  ui i, noOfCaptures, noOfNonCaptures, nMoves;
  BB winningCapture[32], nonCapture[128];
  BB allMoves[128];
  int ambiRank = 0, ambiColumn = 0;

  ASSERT(wToMove == false || wToMove == true);
  ASSERT(ply >= 0 && ply < maxPly);

  /* collecting the data first */
  sq = from(move);
  t = to(move);
  p = piece(move);
  pT = promotedTo(move);
  ASSERT(sq != t);
  ASSERT(p > none && p <= bK);

  /* find ambiguity */
  if (p != wK && p != bK && p != wB && p != bB && p != wP && p != bP)
    { 
      noOfCaptures = captures(wToMove, ply, winningCapture,
			      pos, plyData, others, 0);
      noOfNonCaptures = nonCaptures(wToMove, ply, nonCapture, pos,
				    plyData, others, 0);
      nMoves = 0;
      if (noOfCaptures > 0)
	for (i = 0; i < noOfCaptures; i++)
	  allMoves[nMoves++] = winningCapture[i];
      if (noOfNonCaptures > 0)
	for (i = 0; i < noOfNonCaptures; i++)
	  allMoves[nMoves++] = nonCapture[i];
      for (i = 0; i < nMoves; i++)
	{
	  if (to(allMoves[i]) == t && from(allMoves[i]) != sq &&
	      piece(allMoves[i]) == p)
	    {
	      ambiRank = 1;
	      if (column(from(allMoves[i])) == column(sq)) 
		{
		  ambiRank = 0;
		  ambiColumn = 1;
		}
	    }
	}
    }

  /* the moment of truth */
  /* castling */
  if ((p == wK && sq == E1 && (t == G1 || t == C1)) || 
      (p == bK && sq == E8 && (t == G8 || t == C8)))
    {
      if ((t == G1) || (t == G8)) s += sprintf(s, "O-O");
      else if ((t == C1) || (t == C8)) s += sprintf(s, "O-O-O");
    }
  else
    {
      /* pawn move, just the column */
      if (p == wP || p == bP) *s++ = column(sq)+'a';
      else
	{
	  /* "N", or "Nb", or "N1" */
	  switch (p)
	    {
	    case wN:
	      *s++ = 'N';
	      if (ambiRank) *s++ = column(sq)+'a';
	      if (ambiColumn) *s++ = row(sq)+'1';
	      break;
	    case wB:
	      *s++ = 'B';
	      break;
	    case wR:
	      *s++ = 'R';
	      if (ambiRank) *s++ = column(sq)+'a';
	      if (ambiColumn) *s++ = row(sq)+'1';
	      break;
	    case wQ:
	      *s++ = 'Q';
	      if (ambiRank) *s++ = column(sq)+'a';
	      if (ambiColumn) *s++ = row(sq)+'1';
	      break;
	    case wK:
	      *s++ = 'K';
	      break;
	    case bN:
	      *s++ = 'N';
	      if (ambiRank) *s++ = column(sq)+'a';
	      if (ambiColumn) *s++ = row(sq)+'1';
	      break;
	    case bB:
	      *s++ = 'B';
	      break;
	    case bR:
	      *s++ = 'R';
	      if (ambiRank) *s++ = column(sq)+'a';
	      if (ambiColumn) *s++ = row(sq)+'1';
	      break;
	    case bQ:
	      *s++ = 'Q';
	      if (ambiRank) *s++ = column(sq)+'a';
	      if (ambiColumn) *s++ = row(sq)+'1';
	      break;
	    case bK:
	      *s++ = 'K';
	      break;
	    default:
	      printf("ERROR: undefined piece in san()\n");
	      break;
	    }
	}

      /* capture */
      if ((OneMask[t] & (pos->WBoard | pos->BBoard)) ||
	  /* en-passant */
	  ((p == wP || p == bP) && (OneMask[t] & plyData[ply].EP)))
	{
	  *s++ = 'x';
	  s += sprintf(s, "%c%c", column(t)+'a', row(t)+'1'); 
	}
      else /* non capture */
	{
	  /* pawn move, need only the rank */
	  if (p == wP || p == bP) *s++ = row(t)+'1';
	  /* need both column and rank */
	  else s += sprintf(s, "%c%c", column(t)+'a', row(t)+'1'); 
	}

      /* promotion */
      switch(pT)
	{
	case wN:
	  *s++ = '=';
	  *s++ = 'N';
	  break;
	case wB:
	  *s++ = '=';
	  *s++ = 'B';
	  break;
	case wR:
	  *s++ = '=';
	  *s++ = 'R';
	  break;
	case wQ:
	  *s++ = '=';
	  *s++ = 'Q';
	  break;
	case bN:
	  *s++ = '=';
	  *s++ = 'N';
	  break;
	case bB:
	  *s++ = '=';
	  *s++ = 'B';
	  break;
	case bR:
	  *s++ = '=';
	  *s++ = 'R';
	  break;
	case bQ:
	  *s++ = '=';
	  *s++ = 'Q';
	  break;
	default:
	  break;
	}
    }
  
  /* Do we need a '#' or '+'? */
  makeMove(wToMove, ply+1, move, pos, plyData, others);
  if ((wToMove && 
       (getWAttacker(pos->bKSq, pos, &plyData[ply+1], others) & 2)) ||
      (!wToMove && 
       (getBAttacker(pos->wKSq, pos, &plyData[ply+1], others) & 2)))
    {
      noOfCaptures = captures(wToMove^1, ply+1, winningCapture,
			      pos, plyData, others, 0);
      noOfNonCaptures = nonCaptures(wToMove^1, ply+1, nonCapture,
				    pos, plyData, others, 0);
      if (noOfCaptures+noOfNonCaptures == 0) *s++ = '#';
      else *s++ = '+';
    }
  unmakeMove(wToMove, ply+1, move, pos, plyData);
  *s = '\0';
}


/* Translate move into file_number_file_number notation */
void algebraic(BB move, char *s)
{
  ui sq, t, pT;

  sq = from(move);
  t = to(move);
  pT = promotedTo(move);

  s += sprintf(s, "%c%c%c%c", column(sq)+'a', row(sq)+'1', 
	       column(t)+'a', row(t)+'1');

  switch(pT)
    {
    case wN:
      *s++ = 'n';
      break;
    case wB:
      *s++ = 'b';
      break;
    case wR:
      *s++ = 'r';
      break;
    case wQ:
      *s++ = 'q';
      break;
    case bN:
      *s++ = 'n';
      break;
    case bB:
      *s++ = 'b';
      break;
    case bR:
      *s++ = 'r';
      break;
    case bQ:
      *s++ = 'q';
      break;
    default:
      break;
    }
  *s = '\0';
}


void display(sPos *pos)
{
  ui sq, piece[64];
  BB temp;

  for (sq = A1; sq <= H8; sq++) piece[sq] = 0;

  temp = pos->WPBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = wP;
      temp &= NilMask[sq];
    }

  temp = pos->WNBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = wN;
      temp &= NilMask[sq];
    }

  temp = pos->WBBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = wB;
      temp &= NilMask[sq];
    }

  temp = pos->WRBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = wR;
      temp &= NilMask[sq];
    }

  temp = pos->WQBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = wQ;
      temp &= NilMask[sq];
    }

  piece[pos->wKSq] = wK;

  temp = pos->BPBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = bP;
      temp &= NilMask[sq];
    }

  temp = pos->BNBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = bN;
      temp &= NilMask[sq];
    }

  temp = pos->BBBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = bB;
      temp &= NilMask[sq];
    }

  temp = pos->BRBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = bR;
      temp &= NilMask[sq];
    }

  temp = pos->BQBoard;
  while (temp)
    {
      sq = firstOne(temp);
      piece[sq] = bQ;
      temp &= NilMask[sq];
    }

  piece[pos->bKSq] = bK;

  printf("\n  A B C D E F G H");
  printf("\n8 ");  
  for(sq = A8; sq <= H8; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n7 ");  
  for(sq = A7; sq <= H7; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n6 ");  
  for(sq = A6; sq <= H6; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n5 ");  
  for(sq = A5; sq <= H5; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n4 ");  
  for(sq = A4; sq <= H4; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n3 ");  
  for(sq = A3; sq <= H3; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n2 ");  
  for(sq = A2; sq <= H2; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n1 ");  
  for(sq = A1; sq <= H1; sq++) printf("%c ", boardChars[piece[sq]]);
  printf("\n\n");
}


#ifdef VERIFY
/* From crafty */
void displayBitBoard(BB Board)
{
  union doub 
  {
    char i[8];
    BB d;
  };
  union doub x;
  int i, j;
#if !defined(LINUX)
  ui subs[8] = {7, 6, 5, 4, 3, 2, 1, 0};
#endif

  x.d = Board;
#if !defined(LINUX)
  for(i=7; i>=0; i--) 
    {
      printf("  %2d ", i*8);
      for(j=1; j<=128; j=j<<1)
	{
	  if(x.i[subs[i]] & j) printf("X ");
	  else printf("- ");
	}
      printf("\n");
    }
#else
  for(i=7; i>=0; i--) 
    {
      printf("  %2d ",i*8);
      for(j=1; j<=128; j=j<<1)
	{
	  if(x.i[i] & j) printf("X ");
	  else printf("- ");
	}
      printf("\n");
    }
#endif /* LINUX */
      printf("\n");
}


void displayBitRank(ui bitRank)
{
  ui i;

  ASSERT(bitRank >= 0);

  for(i=1; i<=128; i=i<<1)
    {
      if(bitRank & i) printf("X ");
      else printf("- ");
    }
  printf("\n");
}
#endif /* VERIFY */


/*
 * Set up a board position based on FEN notation.  For example for command:
 *         setboard 3r2k1/p4ppp/1q6/8/8/2R1P3/P3QPPP/6K1 b - - 0 1
 */
void setBoard(char s1[LETTERS], char s2, char* s3, char* s4, char* s5,
	      char* s6, ui *wToMove, sPos *pos, sPlyData *thisPlyData, 
	      sOthers *others)
{
  ui i, c, r, ep;
  ui sqstart, sq, piece[64];

  ASSERT(*wToMove == false || *wToMove == true);

  /* Empty board */
  for (sq = A1; sq <= H8; sq++) piece[sq] = none;

  /* Piece positions */
  sq = sqstart = A8;
  i = 0;
  while (s1[i++])
    {
      switch (s1[i-1])
	{
	case 'p':
	  piece[sq] = bP;
	  sq++;
	  break;
	case 'n':
	  piece[sq] = bN;
	  sq++;
	  break;
	case 'b':
	  piece[sq] = bB;
	  sq++;
	  break;
	case 'r':
	  piece[sq] = bR;
	  sq++;
	  break;
	case 'q':
	  piece[sq] = bQ;
	  sq++;
	  break;
	case 'k':
	  piece[sq] = bK;
	  sq++;
	  break;
	case 'P':
	  piece[sq] = wP;
	  sq++;
	  break;
	case 'N':
	  piece[sq] = wN;
	  sq++;
	  break;
	case 'B':
	  piece[sq] = wB;
	  sq++;
	  break;
	case 'R':
	  piece[sq] = wR;
	  sq++;
	  break;
	case 'Q':
	  piece[sq] = wQ;
	  sq++;
	  break;
	case 'K':
	  piece[sq] = wK;
	  sq++;
	  break;
	case '1':
	  sq += 1;
	  break;
	case '2':
	  sq += 2;
	  break;
	case '3':
	  sq += 3;
	  break;
	case '4':
	  sq += 4;
	  break;
	case '5':
	  sq += 5;
	  break;
	case '6':
	  sq += 6;
	  break;
	case '7':
	  sq += 7;
	  break;
	case '8':
	  sq += 8;
	  break;
	case '/':
	  sqstart -= 8;
	  sq = sqstart;
	  break;
	default:
	  fprintf (stderr, "Strange character in FEN\n");
	  return;
	}
    }

#ifdef VERIFY
  for (sq = A1; sq <= H8; sq++) thisPlyData->piece[sq] = piece[sq];
#endif

  /* castling information */
  thisPlyData->wCanCastle = thisPlyData->bCanCastle = 0;
  for (i = 0; i < (ui) strlen(s3); i++)
    {
      if (s3[i] == 'K') thisPlyData->wCanCastle |= 1;
      if (s3[i] == 'Q') thisPlyData->wCanCastle |= 2;
      if (s3[i] == 'k') thisPlyData->bCanCastle |= 1;
      if (s3[i] == 'q') thisPlyData->bCanCastle |= 2;
    }

  /* en-passant square */
  if (s4[0] == '-' || s4[0] == '0')
    {
      thisPlyData->EP = 0;      
    }
  else
    {
      c = (ui) s4[0] - (ui) 'a';
      r = (ui) s4[1] - (ui) '1';
      ep = r*8 + c;
      thisPlyData->EP = OneMask[ep];
    }

  /* This has to be done before settings side to move, etc. */
  finalizeInit(piece, pos, thisPlyData, others);
  display(pos);

  /* side to move.  "go" switch pos->iAm */
  if (s2 == 'w') 
    {
      *wToMove = true;
      pos->iAm = black;
    }
  else if (s2 == 'b') 
    {
      *wToMove = false;
      pos->iAm = white;
    }  

  /* 50 move count */
  thisPlyData->count50Moves = (ui) atoi(s5);

  /* move number */
  pos->moveNumber = (ui) atoi(s6);
  if (s2 == 'w' && pos->moveNumber > 0) pos->moveNumber--;
}


/*
 * Return the FEN string of a position.
 */
void getFEN(ui wToMove, sPos *pos, sPlyData *thisPlyData, char *fen)
{
  int i, j, blanks;
  ui sq, piece[64];
  BB Temp;
  ui ep;

  ASSERT(wToMove == false || wToMove == true);

  for (i = A1; i <= H8; i++) piece[i] = 0;
  
  /* white pawns */
  Temp = pos->WPBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = wP;
      Temp &= NilMask[sq];
    }      
  
  /* white knights */
  Temp = pos->WNBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = wN;
      Temp &= NilMask[sq];
    }      
  
  /* white bishops */
  Temp = pos->WBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = wB;
      Temp &= NilMask[sq];
    }      
  
  /* white rooks */
  Temp = pos->WRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = wR;
      Temp &= NilMask[sq];
    }      
  
  /* white queens */
  Temp = pos->WQBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = wQ;
      Temp &= NilMask[sq];
    }      
  
  /* white king */
  piece[pos->wKSq] = wK;
  
  /* black pawns */
  Temp = pos->BPBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = bP;
      Temp &= NilMask[sq];
    }      
  
  /* black knights */
  Temp = pos->BNBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = bN;
      Temp &= NilMask[sq];
    }      
  
  /* black bishops */
  Temp = pos->BBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = bB;
      Temp &= NilMask[sq];
    }      
  
  /* black rooks */
  Temp = pos->BRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = bR;
      Temp &= NilMask[sq];
    }      
  
  /* black queens */
  Temp = pos->BQBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[sq] = bQ;
      Temp &= NilMask[sq];
    }      
  
  /* black king */
  piece[pos->bKSq] = bK;

  /* the pieces */
  i = 0;
  for (j = A8; j >= A1; j -= 8)
    {
      blanks = 0;
      for (sq = (ui) j; sq <= (ui) j+7; sq++)
	{
	  switch(piece[sq])
	    {
	    case wP:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'P';
	      break;
	    case wN:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'N';
	      break;
	    case wB:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'B';
	      break;
	    case wR:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'R';
	      break;
	    case wQ:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'Q';
	      break;
	    case wK:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'K';
	      break;
	    case bP:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'p';
	      break;
	    case bN:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'n';
	      break;
	    case bB:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'b';
	      break;
	    case bR:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'r';
	      break;
	    case bQ:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'q';
	      break;
	    case bK:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'k';
	      break;
	    default:
	      blanks++;
	      break;
	    }
	}
      if (blanks != 0)
	{
	  fen[i++] = blanks+'0';
	  blanks = 0;
	}
      if (j != A1) fen[i++] = '/';
      else fen[i++] = ' ';
    }

  /* side to move */
  if (wToMove) fen[i++] = 'w';
  else fen[i++] = 'b';
  fen[i++] = ' ';

  /* castling information */
  if (thisPlyData->wCanCastle <= 0 && thisPlyData->bCanCastle <= 0)
    fen[i++] = '-';
  else
    {
      if (thisPlyData->wCanCastle > 0)
	{
	  if (thisPlyData->wCanCastle & 1) fen[i++] = 'K';
	  if (thisPlyData->wCanCastle & 2) fen[i++] = 'Q';
	}
      if (thisPlyData->bCanCastle > 0)
	{
	  if (thisPlyData->bCanCastle & 1) fen[i++] = 'k';
	  if (thisPlyData->bCanCastle & 2) fen[i++] = 'q';
	}
    }
  fen[i++] = ' ';

  /* en-passant square */
  if (thisPlyData->EP)
    {
      ep = firstOne(thisPlyData->EP);
      fen[i++] = column(ep) + 'a';
      fen[i++] = row(ep) + '1';
    }
  else fen[i++] = '-';
  fen[i++] = ' ';

  /* 50 move count */
  if (thisPlyData->count50Moves > 9) 
    fen[i++] = (int)(thisPlyData->count50Moves/10) + '0';
  fen[i++] = thisPlyData->count50Moves%10 + '0';
  fen[i++] = ' ';

  /* move number */
  if (pos->moveNumber > 9) 
    fen[i++] = (int)(pos->moveNumber/10) + '0';
  fen[i++] = pos->moveNumber%10 + '0';

  /* end the string */
  fen[i] = '\0';
}


/*
 * Return the reverse FEN string of a position.
 */
void reverse(ui *wToMove, sPos *pos, sPlyData *thisPlyData, sOthers *others,
	     char *fen)
{
  int i, j, blanks;
  ui sq, piece[64];
  BB Temp;
  ui ep;
  BB *pWAttacker = (BB *) thisPlyData->wAttacker;
  BB *pBAttacker = (BB *) thisPlyData->bAttacker;

  ASSERT(*wToMove == false || *wToMove == true);

  for (i = A1; i <= H8; i++) piece[i] = 0;

  /* Attack tables */
  pWAttacker[0] =   pWAttacker[1] =   pWAttacker[2] =   pWAttacker[3] = 
    pWAttacker[4] =   pWAttacker[5] =   pWAttacker[6] =   pWAttacker[7] = 0; 
  pBAttacker[0] =   pBAttacker[1] =   pBAttacker[2] =   pBAttacker[3] = 
    pBAttacker[4] =   pBAttacker[5] =   pBAttacker[6] =   pBAttacker[7] = 0; 

  /* white pawns */
  Temp = pos->WPBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = bP;
      Temp &= NilMask[sq];
    }      
  
  /* white knights */
  Temp = pos->WNBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = bN;
      Temp &= NilMask[sq];
    }      
  
  /* white bishops */
  Temp = pos->WBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = bB;
      Temp &= NilMask[sq];
    }      
  
  /* white rooks */
  Temp = pos->WRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = bR;
      Temp &= NilMask[sq];
    }      
  
  /* white queens */
  Temp = pos->WQBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = bQ;
      Temp &= NilMask[sq];
    }      
  
  /* white king */
  piece[reverseSq[pos->wKSq]] = bK;
  
  /* black pawns */
  Temp = pos->BPBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = wP;
      Temp &= NilMask[sq];
    }      
  
  /* black knights */
  Temp = pos->BNBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = wN;
      Temp &= NilMask[sq];
    }      
  
  /* black bishops */
  Temp = pos->BBBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = wB;
      Temp &= NilMask[sq];
    }      
  
  /* black rooks */
  Temp = pos->BRBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = wR;
      Temp &= NilMask[sq];
    }      
  
  /* black queens */
  Temp = pos->BQBoard;
  while (Temp)
    {
      sq = firstOne(Temp);
      piece[reverseSq[sq]] = wQ;
      Temp &= NilMask[sq];
    }      
  
  /* black king */
  piece[reverseSq[pos->bKSq]] = wK;

  /* the pieces */
  i = 0;
  for (j = A8; j >= A1; j -= 8)
    {
      blanks = 0;
      for (sq = (ui) j; sq <= (ui) j+7; sq++)
	{
	  switch(piece[sq])
	    {
	    case wP:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'P';
	      break;
	    case wN:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'N';
	      break;
	    case wB:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'B';
	      break;
	    case wR:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'R';
	      break;
	    case wQ:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'Q';
	      break;
	    case wK:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'K';
	      break;
	    case bP:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'p';
	      break;
	    case bN:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'n';
	      break;
	    case bB:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'b';
	      break;
	    case bR:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'r';
	      break;
	    case bQ:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'q';
	      break;
	    case bK:
	      if (blanks != 0)
		{
		  fen[i++] = blanks+'0';
		  blanks = 0;
		}
	      fen[i++] = 'k';
	      break;
	    default:
	      blanks++;
	      break;
	    }
	}
      if (blanks != 0)
	{
	  fen[i++] = blanks+'0';
	  blanks = 0;
	}
      if (j != A1) fen[i++] = '/';
      else fen[i++] = ' ';
    }

  /* side to move */
  if (*wToMove) fen[i++] = 'b';
  else fen[i++] = 'w';
  *wToMove ^= 1;
  pos->iAm ^= 1;
  fen[i++] = ' ';

  /* castling information */
  if (thisPlyData->wCanCastle <= 0 && thisPlyData->bCanCastle <= 0)
    fen[i++] = '-';
  else
    {
      if (thisPlyData->bCanCastle > 0)
	{
	  if (thisPlyData->bCanCastle & 1) fen[i++] = 'K';
	  if (thisPlyData->bCanCastle & 2) fen[i++] = 'Q';
	}
      if (thisPlyData->wCanCastle > 0)
	{
	  if (thisPlyData->wCanCastle & 1) fen[i++] = 'k';
	  if (thisPlyData->wCanCastle & 2) fen[i++] = 'q';
	}
    }
  fen[i++] = ' ';

  /* en-passant square */
  if (thisPlyData->EP)
    {
      ep = firstOne(thisPlyData->EP);
      fen[i++] = column(reverseSq[ep]) + 'a';
      fen[i++] = row(reverseSq[ep]) + '1';
    }
  else fen[i++] = '-';
  fen[i++] = ' ';

  /* 50 move count */
  if (thisPlyData->count50Moves > 9) 
    fen[i++] = (int)(thisPlyData->count50Moves/10) + '0';
  fen[i++] = thisPlyData->count50Moves%10 + '0';
  fen[i++] = ' ';

  /* move number */
  if (pos->moveNumber > 9) 
    fen[i++] = (int)(pos->moveNumber/10) + '0';
  fen[i++] = pos->moveNumber%10 + '0';

  /* end the string */
  fen[i] = '\0';

  /* This has to be at the end, after fen[] is written correctly. */
  finalizeInit(piece, pos, thisPlyData, others);
  display(pos);
}


/*
 * Find the best move in pos between indeces 
 * head and tail.  Swap the move with highest score to the head.
 */
void pick(ui head, ui tail, BB *moveList)
{
  ui i, maxIndex;
  int max;
  BB temp;

  ASSERT(head >= 0);
  ASSERT(tail >= 0);

  maxIndex = head;
  max = score(moveList[head]);

  for (i = head+1; i < tail; i++)
    {
      if (max < score(moveList[i]))
	{
	  maxIndex = i;
	  max = score(moveList[i]);
	}
    }

  if (maxIndex != head)
    {
      temp = moveList[head];
      moveList[head] = moveList[maxIndex];
      moveList[maxIndex] = temp;
    }
}


/* "borrowed" from Crafty */
ui firstOne(BB arg1)
{
  if (arg1 & 65535)
    return (fOne[arg1 & 65535]);
  if ((arg1 >> 16) & 65535)
    return (fOne[(arg1 >> 16) & 65535] + 16);
  if ((arg1 >> 32) & 65535)
    return (fOne[(arg1 >> 32) & 65535] + 32);
  return (fOne[arg1 >> 48] + 48);
}


/* "borrowed" from Crafty */
ui lastOne(BB arg1)
{
  if (!arg1) return 64;
  if (arg1 >> 48)
    return (63 - lOne[arg1 >> 48]);
  if ((arg1 >> 32) & 65535)
    return (47 - lOne[(arg1 >> 32) & 65535]);
  if ((arg1 >> 16) & 65535)
    return (31 - lOne[(arg1 >> 16) & 65535]);
  return (15 - lOne[arg1 & 65535]);
}


ui bitCount(register BB a)
/* Count the number of one bits in an unsigned long long variable */
{
  register int c=0;

#ifdef HYATT
  while(a) {
    c++;
    a &= a - 1;
  }
  return(c);
#endif

  /* Bruce Moreland */
  if (!a) return (0);
  else if (!(a &= a - 1)) return (1);
  else if (!(a &= a - 1)) return (2);
  else
    for (c = 3; ; c++)
      if (!(a &= a - 1)) return (c);
}


void closing(FILE *fd, sPos *pos, char *result, char *comment)
{
#if !defined TOURNAMENT
  FILE *fd1;
#endif

  pos->Flags |= OneMask[4]; 
#ifdef _REENTRANT
  pthread_mutex_lock(&goMutex);
#endif
  if (go & goPonder) go = goRest;
#ifdef _REENTRANT
  pthread_mutex_unlock(&goMutex);
#endif

  if (pos->iAm == white)
    {
      if (strstr(result, "1-0"))
	fprintf(fd, "Result:  win w %5u %3u moves %s\n", 
		pos->openingNo, (pos->moveNumber+1), comment);
      else if (strstr(result, "0-1"))
	fprintf(fd, "Result: lost w %5u %3u moves %s\n", 
		pos->openingNo, (pos->moveNumber+1), comment);
      else
	fprintf(fd, "Result: draw w %5u %3u moves %s\n", 
		pos->openingNo, (pos->moveNumber+1), comment);
    }
  else
    {
      if (strstr(result, "1-0"))
	fprintf(fd, "Result: lost b %5u %3u moves %s\n", 
		pos->openingNo, (pos->moveNumber+1), comment);
      else if (strstr(result, "0-1"))
	fprintf(fd, "Result:  win b %5u %3u moves %s\n", 
		pos->openingNo, (pos->moveNumber+1), comment);
      else
	fprintf(fd, "Result: draw b %5u %3u moves %s\n", 
		pos->openingNo, (pos->moveNumber+1), comment);
    }
  fflush(fd);

#if !defined TOURNAMENT
  /* write down the used openingNo */
  if (pos->moveNumber > 9)
    {
      if (pos->iAm == white)
	{
	  fd1 = fopen("usedWOpenings", "a+");
	  fprintf(fd1, "%d\n", pos->openingNo);
	  fclose(fd1);
	}
      else
	{
	  fd1 = fopen("usedBOpenings", "a+");
	  fprintf(fd1, "%d\n", pos->openingNo);
	  fclose(fd1);
	}
    }
#endif
}


void exiting(FILE *fd, sHash **hash)
{
  free(hash[white]);
  free(hash[black]);
  fclose(fd);
  exit(0);
}
