/* values.h */
/* Last modified: July 22, 2007 */


#define version "070722"


/* logic and characters */
#define true 1
#define false 0
#define newLine 10

/* maximum number of words and letters sent to runCommand() */
#define WORDS 12
#define LETTERS 80

/* except mateScore, the others are for game listings */
#define infinity     32768
#define mateScore    30000
#define almostScore  29000
#define drawScore    25000
#define forcedScore  20000
#define onlyScore    15000
#define openingScore 10000


/* piece values */
#define pValue 100
#define nValue 300
#define bValue 301         /* mainly so that we can tell bishop from knight */
#define rValue 500
#define qValue 900
#define endGameValue 1400  /* rValue + qValue */
#define middleGameValue 2227  /* value of all pieces minus queen */
#define kValue mateScore


/* search related */
#define goProcess 1
#define goRest    2
#define goSearch  4
#define goPonder  8
#define maxSearchBudget 1e14        /* in microseconds */
#define ponderingBudget 1e15        /* in microseconds */
#define NPMS 0.9  /* estimated nodes per microsecond */
#define maxMove 256
#define minPly 4
#define maxPly 64
#define easyScore 200
#define easyFactor 3
#define halfWindow 40
#define drawCountLimit 25
#define drawOfferRepeat 5
#define drawMargin 25
#define resignCountLimit 5
#define resignValue 900
#define pruneMargin 300
#define allNode -1
#define PVNode 0
#define cutNode 1
#define nonEndgame 0
#define endgame 1
#define averageBFactor 4.65        /* collected between January and May 2007 */
#define stableValue 40
#define iidDepth 3
#define iidReduction 2
#define pvExt0 25
#define checkExt0 100
#define onlyReplyExt0 75
#define mateThreatExt0 75
#define parrotExt0 38
#define pawn7thExt0 75
#define lateMoveThreshold 4


/* eval related */
#define BONUS00    3 /* knight mobility bonus per possible move */
#define BONUS01   50 /* knight unassailable at D5/E5/D4/E4 */
#define BONUS02   13 /* knight protected at D5/E5/D4/E4 */
#define BONUS03  100 /* knight unassailable at D6/E6/D3/E3 */
#define BONUS04   25 /* knight protected at D6/E6/D3/E3 */
#define BONUS05    5 /* bishop mobility bonus per possible move */
#define BONUS06   10 /* semi open file */
#define BONUS07   35 /* rook on 7th rank */
#define BONUS08   10 /* queen on 7th rank */
#define BONUS09    3 /* king good neighborhood bonus per friendly piece */
#define BONUS10   40 /* rook behind passed pawn */
#define BONUS11   10 /* pawn pair */
#define BONUS12   25 /* both side pawns, bishop versus knight */
#define BONUS13   15 /* winning in end game, king chases opponent king */
#define BONUS14   10 /* king sheltered from pesky rook */
#define BONUS15   20 /* open file */
#define BONUS16   10 /* ahead in development */
#define BONUS17    2 /* pawn storm in general */
#define BONUS18    2 /* pawn storm toward enemy king */

#define PENALTY00    25 /* piled pawn */
#define PENALTY01    20 /* pawn isolani */
#define PENALTY02    30 /* bishop blocking D2/E2/D7/E7 */
#define PENALTY03    25 /* rook obstructed by king */
#define PENALTY04    25 /* rook obstructed by pawn */
#define PENALTY06     5 /* missing pawn 1 in front of king */
#define PENALTY07     5 /* missing pawn 2 in front of king */
#define PENALTY08    20 /* missing pawn 3 in front of king */
#define PENALTY09     3 /* missing pawn 1 outside king */
#define PENALTY10     3 /* missing pawn 2 outside king */
#define PENALTY11    12 /* missing pawn 3 outside king */
#define PENALTY12     2 /* missing pawn 1 inside king */
#define PENALTY13     2 /* missing pawn 2 inside king */
#define PENALTY14     8 /* missing pawn 3 inside king */
#define PENALTY15     5 /* B2/G2/B7/G7 hole */
#define PENALTY16    20 /* losing castling right */
#define PENALTY17    30 /* basis for "mayor attack" */
#define PENALTY18    10 /* king in open file */
#define PENALTY19     6 /* outside king is open file */
#define PENALTY20     4 /* inside king is open file */
#define PENALTY21    70 /* double bonus reduction due to passed pawn */


/* For SAN notation */
#define rxx "12345678"
#define cxx "abcdefgh"
#define sanChars   " PNBRQKPNBRQK"
#define boardChars "-FNBRQKpnbrqk"


/* "E4" is easier to be remembered than "28" */
#define A1  0
#define B1  1
#define C1  2
#define D1  3
#define E1  4
#define F1  5
#define G1  6
#define H1  7
#define A2  8
#define B2  9
#define C2 10
#define D2 11
#define E2 12
#define F2 13
#define G2 14
#define H2 15
#define A3 16
#define B3 17
#define C3 18
#define D3 19
#define E3 20
#define F3 21
#define G3 22
#define H3 23
#define A4 24
#define B4 25
#define C4 26
#define D4 27
#define E4 28
#define F4 29
#define G4 30
#define H4 31
#define A5 32
#define B5 33
#define C5 34
#define D5 35
#define E5 36
#define F5 37
#define G5 38
#define H5 39
#define A6 40
#define B6 41
#define C6 42
#define D6 43
#define E6 44
#define F6 45
#define G6 46
#define H6 47
#define A7 48
#define B7 49
#define C7 50
#define D7 51
#define E7 52
#define F7 53
#define G7 54
#define H7 55
#define A8 56
#define B8 57
#define C8 58
#define D8 59
#define E8 60
#define F8 61
#define G8 62
#define H8 63

/* colors */
#define white 0
#define black 1

/* pieces */
#define none 0
#define wP 1
#define wN 2
#define wB 3
#define wR 4
#define wQ 5
#define wK 6
#define bP 7
#define bN 8
#define bB 9
#define bR 10
#define bQ 11
#define bK 12


/* flags */
#define flagDraw3  1
#define flagDraw13 13
#define flagDraw50 2


/* hash tables */
#define bookSize 131071        /* 2^17 - 1*/
#define EXACT         1
#define FAIL_LOW      2
#define FAIL_HIGH     4
