/* variables.h */
/* Last modified: June 25, 2007 */


#ifdef WINDOWS
#include <time.h>
#else
#include <sys/time.h>
#endif


/* 16 bytes */
struct BookTable
{
  BB move;
  ui key;
  unsigned short no;
  unsigned short count;
};
static struct BookTable wBookTable[(bookSize+1)], bBookTable[(bookSize+1)];


/* 32 bytes */
struct TimeCheck
{
  double timeBudget;        /* in micro seconds */
#ifdef WINDOWS
  double clock0;
#else
  BB clock0;
#endif
  BB nodeInc, nodeCheckPoint;
};


/* 32 bytes */
struct HashTable
{
  BB key;
  BB stuffs;
  BB keySave;
  BB stuffsSave;
};


struct PV
{
  BB move[maxPly];
  ui length;
};


/* 424 bytes */
struct PlyData
{
#ifdef VERIFY
  ui piece[64];
#endif
  /****************************************
   * Attack tables based on Rebel's       *
   * bit    0: not calculated/calculated  *
   * bit    1: not attacked/attacked      *
   * bit    2: not used                   * 
   * bit    3: attack by pawn             *
   * bit    4: attack by knight or bishop *
   * bit    5: attack by rook             *
   * bit    6: attack by queen            *
   * bit    7: attack by king             *
   ****************************************/
  unsigned char wAttacker[64], bAttacker[64];
  BB key;
  BB EP;
  int wCanCastle, bCanCastle;  /* 01  : can castle short  
				* 02  : can castle long 
				* 00  : lost castling right
				* -ply: castled on ply ply
				*/
  ui count50Moves;
};


/* 708 bytes - 704 bytes will be ideal */
struct Position
{
  /* WARNING: others->moveNumber starts from 0, not 1 like the real move 
   * number.  It needs to start from 0 because it's used as an index of
   * gameListing[].
   */
  ui failBook, openingNo;
  ui rOpening;
  ui moveNumber;
  ui iteration, fastIteration, searchDepth;
  ui movesPerTC;
  ui wKSq, bKSq;
  ui iAm;
  ui repHead;
  ui EGTBlimit;  /* maximum pieces in our endgame tablebase */
  ui hashSize, cacheSize;
  int rootAlpha, rootBeta;
  ui rootMaxPieceScore;
  ui deltaRating;
  ui wScore;
  ui pvExt;
  ui checkExt;
  ui onlyReplyExt;
  ui mateThreatExt;
  ui parrotExt;
  ui pawn7thExt;

  ui wPCount[8], bPCount[8];
  ui killerFrom[maxPly], killerTo[maxPly];

  double timeLeft0, timeLeft, oTimeLeft, timeInc;  /* in centiSecond */

  /* Put all true/false global values into this variable
   *        bit        description
   *          0        xboard
   *          1        forced
   *          2        game is already listed
   *          3        randomize
   *          4        search terminated
   *          5        program terminated
   *          6        book disabled
   *          7        search almost terminated
   *          8        drawish
   *          9        pawn king square rule was used
   *         10        timeReserved mode
   *         11        ponder on
   */
  BB Flags;

  BB qNodes, sNodes;
  BB WBoard, BBoard, WPBoard, BPBoard, WRBoard, BRBoard, 
    WNBoard, BNBoard, WBBoard, BBBoard, WQBoard, BQBoard;
  BB RotatedBoard90, RotatedBoard45l, RotatedBoard45r;
  BB repKey[102];

  char egtbDir[80];
  char opponentName[80];
  char posID[80], solution[80], myMove[8];
  /* for debugging: "nodeNumber_ply_hashKeyBeforeTheMove_wToMove_sanMove" */
  char gameID[80], fullDate[16];
};


/* 28792 bytes - crazy and messy */
struct Others
{
 /* all the bitboards */
  BB QueenSide, FarQueenSide, FNQueenSide, FSQueenSide;
  BB KingSide, FarKingSide, FNKingSide, FSKingSide;
  BB Edge, InnerEdge, Dark, Light;

  BB Row[8], Column[8];

  BB Up1Of[64], Up7Of[64], Up8Of[64], Up9Of[64];
  BB Down1Of[64], Down7Of[64], Down8Of[64], Down9Of[64];
  BB SideOf[64], Up8SideOf[64], Down8SideOf[64];

  /* 3072 bytes */
  BB OneUp[64], TwoUp[64], OneDown[64], TwoDown[64], OneSide[64], TwoSide[64];

  /* 3072 bytes */
  BB WPCaptures[64] , BPCaptures[64];
  BB NMoves[64], BMoves[64], KMoves[64], KHood[64];

  /* square value tables */
  /* 3584 bytes */
  int wPSqvalue[64], wNSqvalue[64], wBSqvalue[64];
  int wRSqvalue[64], wQSqvalue[64], wKSqvalue[64];
  int bPSqvalue[64], bNSqvalue[64], bBSqvalue[64];
  int bRSqvalue[64], bQSqvalue[64], bKSqvalue[64];
  int wPassedPSqvalue[64], bPassedPSqvalue[64];

  /* all the hash keys */
  /* 7680 bytes */
  BB wPKey[64], wNKey[64], wBKey[64], wRKey[64], wQKey[64], wKKey[64];
  BB bPKey[64], bNKey[64], bBKey[64], bRKey[64], bQKey[64], bKKey[64];
  BB ePKey[64];

  BB wCanCastleKey[4], bCanCastleKey[4];

  char direction[64][64];
};
